import random
from typing import List

import numpy as np

from ballet.util import asarray2d
from ballet.util.log import logger
from ballet.util.testing import seeded
from ballet.validation.base import FeatureAcceptanceMixin, FeatureAccepter
from ballet.validation.common import (
    RandomFeaturePerformanceEvaluator, load_spec,)
from ballet.validation.entropy import (
    estimate_conditional_information, estimate_mutual_information,)
from ballet.validation.gfssf import (
    GFSSFIterationInfo, GFSSFPerformanceEvaluator, _compute_lmbdas,
    _compute_threshold, _concat_datasets,)


class NeverAccepter(FeatureAccepter):

    def judge(self):
        logger.info(f'Judging feature using {self}')
        return False


class RandomAccepter(FeatureAcceptanceMixin,
                     RandomFeaturePerformanceEvaluator):

    def judge(self):
        """Accept feature with probability p"""
        logger.info(f'Judging feature using {self}')
        with seeded(self.seed):
            return random.uniform(0, 1) < self.p


class AlwaysAccepter(FeatureAccepter):
    def judge(self):
        logger.info(f'Judging feature using {self}')
        return True


class GFSSFAccepter(FeatureAcceptanceMixin, GFSSFPerformanceEvaluator):

    def judge(self):
        """Judge feature acceptance using GFSSF

        Uses lines 1-8 of agGFSSF where we do not remove accepted but
        redundant features on line 8.
        """

        logger.info(f'Judging feature using {self}')

        feature_df_map = self._get_feature_df_map()

        candidate_df = feature_df_map[self.candidate_feature]
        n_samples, n_candidate_cols = candidate_df.shape

        lmbda_1, lmbda_2 = _compute_lmbdas(
            self.lmbda_1, self.lmbda_2, feature_df_map)

        logger.debug(
            f'Recomputed lambda_1={lmbda_1:0.3e}, lambda_2={lmbda_2:0.3e}')

        info = []

        omit_in_test = [None, *self.features]
        n_omit = len(omit_in_test)
        for i, omitted_feature in enumerate(omit_in_test):

            z = _concat_datasets(
                feature_df_map, n_samples,
                omit=[self.candidate_feature, omitted_feature])

            # Calculate CMI of candidate feature
            cmi = estimate_conditional_information(candidate_df, self.y_val, z)

            if omitted_feature is not None:
                omit_df = feature_df_map[omitted_feature]
                _, n_omit_cols = omit_df.shape

                # Calculate CMI of omitted feature
                cmi_omit = estimate_conditional_information(
                    omit_df, self.y_val, z)
            else:
                cmi_omit = 0
                n_omit_cols = 0

                # want to log to INFO only the case of I(Z|Y;X) where X is the
                # entire feature matrix, i.e. no omitted features.
                logger.info(f'I(feature ; target | existing_features) = {cmi}')

            statistic = cmi - cmi_omit
            threshold = _compute_threshold(
                lmbda_1, lmbda_2, n_candidate_cols, n_omit_cols)
            delta = statistic - threshold

            if delta >= 0:
                omitted_source = getattr(omitted_feature, 'source', 'None')
                logger.debug(
                    f'Succeeded while omitting feature: {omitted_source}')

                return True
            else:
                iteration_info = GFSSFIterationInfo(
                    i=i,
                    n_samples=n_samples,
                    candidate_feature=self.candidate_feature,
                    candidate_cols=n_candidate_cols,
                    candidate_cmi=cmi,
                    omitted_feature=omitted_feature,
                    omitted_cols=n_omit_cols,
                    omitted_cmi=cmi_omit,
                    statistic=statistic,
                    threshold=threshold,
                    delta=delta,
                )
                info.append(iteration_info)
                logger.debug(
                    f'Completed iteration {i}/{n_omit}: {iteration_info}')

        info_closest = max(info, key=lambda x: x.delta)
        cmi_closest = info_closest.candidate_cmi
        omitted_cmi_closest = info_closest.omitted_cmi
        statistic_closest = info_closest.statistic
        threshold_closest = info_closest.threshold
        logger.info(
            f'Rejected feature: best marginal conditional mutual information was not greater than threshold ({cmi_closest:0.3e} - {omitted_cmi_closest:0.3e} = {statistic_closest:0.3e}, vs needed {threshold_closest:0.3e}).')  # noqa

        return False


class VarianceThresholdAccepter(FeatureAccepter):
    """Accept features with variance above a threshold

    Args:
        threshold: variance threshold
    """

    def __init__(self, *args, threshold=0.05):
        super().__init__(*args)
        self.threshold = threshold

    def judge(self):
        logger.info(f'Judging feature using {self}')
        z = (
            self.candidate_feature
            .as_feature_engineering_pipeline()
            .fit(self.X_df, y=self.y_df)
            .transform(self.X_df_val)
        )
        var = np.var(z, axis=0)
        delta = var - self.threshold
        outcome = np.all(delta > 0)
        logger.info(
            f'Feature variance is {var} vs. threshold {self.threshold} '
            f'({delta} above threshold)')
        return outcome

    def __str__(self):
        return f'{super().__str__()}: threshold={self.threshold}'


class MutualInformationAccepter(FeatureAccepter):
    """Accept features with mutual information with the target above a threshold

    Args:
        threshold: mutual information threshold
        handle_nan_targets: one of ``'fail'`` or ``'ignore'``, whether to
            fail validation if NaN-valued targets are discovered or to drop
            those rows in calculation of the mutual information score
    """

    def __init__(self, *args, threshold=0.05, handle_nan_targets='fail'):
        super().__init__(*args)
        self.threshold = threshold
        self.handle_nan_targets = handle_nan_targets

    def judge(self):
        logger.info(f'Judging feature using {self}')
        z = (
            self.candidate_feature
            .as_feature_engineering_pipeline()
            .fit(self.X_df, y=self.y_df)
            .transform(self.X_df_val)
        )
        y = self.y_val
        z, y = asarray2d(z), asarray2d(y)
        z, y = self._handle_nans(z, y)
        if z is None and y is None:
            # nans were found and handle_nan_targets == 'fail'
            return False
        mi = estimate_mutual_information(z, y)
        delta = mi - self.threshold
        outcome = delta > 0
        logger.info(
            f'Mutual information with target I(Z;Y) is {mi} vs. '
            f'threshold {self.threshold} ({delta} above threshold)')
        return outcome

    def _handle_nans(self, z, y):
        nans = np.any(np.isnan(y), 1)  # whether there are any nans in this row
        if np.any(nans):
            if self.handle_nan_targets == 'fail':
                return None, None  # hack
            elif self.handle_nan_targets == 'ignore':
                z = z[~nans, :]
                y = y[~nans, :]
            else:
                raise ValueError(
                    'Invalid value for handle_nan_targets: '
                    f'{self.handle_nan_targets}'
                )

        return z, y

    def __str__(self):
        return f'{super().__str__()}: threshold={self.threshold}'


class CompoundAccepter(FeatureAccepter):
    """A compound accepter that runs a list of individual accepters

    An accepter spec is just a simple serialization of a class and its kwargs::

        name: ballet.validation.feature_acceptance.validator.CompoundAccepter
        params:
          agg: any
          specs:
            - name: ballet.validation.feature_acceptance.validator.VarianceThresholdAccepter
              params:
                threshold: 0.1
            - name: ballet.validation.feature_acceptance.validator.MutualInformationAccepter
              params:
                threshold: 0.1

    Args:
        agg: one of ``'all'`` or ``'any'``; whether to accept if all
            underlying accepters accept or if any accepter accepts.
        specs: list of dicts of accepter specs
    """  # noqa

    def __init__(self, *args, agg='all', specs: List[dict] = []):
        super().__init__(*args)
        self._agg = agg
        self._specs = specs
        if not self._specs:
            raise ValueError('Missing list of accepter specs!')
        self.accepters = []
        for spec in self._specs:
            cls, params = load_spec(spec)
            self.accepters.append(cls(*args, **params))
        if self._agg == 'all':
            self.agg = all
        elif self._agg == 'any':
            self.agg = any
        else:
            raise ValueError(
                f'Unsupported value for parameter agg: {self._agg}')

    def judge(self):
        logger.info(f'Judging feature using {self}')
        outcomes = {
            accepter.__class__.__name__: accepter.judge()
            for accepter in self.accepters
        }
        logger.debug(f'Got outcomes {outcomes!r} from underlying accepters')
        return self.agg(outcomes.values())

    def __str__(self):
        accepter_str = ', '.join(str(accepter) for accepter in self.accepters)
        return f'{super().__str__()}: '\
            f'agg={self._agg}, accepters=({accepter_str})'



import pytest


@pytest.mark.xfail
def test_validate_feature_api():
    raise NotImplementedError


@pytest.mark.xfail
def test_validate_feature_acceptance():
    raise NotImplementedError


@pytest.mark.xfail
def test_discover():
    pass



import re

import pytest
from pkg_resources import working_set


@pytest.mark.slow
def test_quickstart_install(quickstart, virtualenv):
    """Test that we can install the project resulting from quickstart

    Note that virtualenv.install_package is just bogus and should be avoided.
    """
    d = quickstart.tempdir.joinpath(quickstart.project_slug).resolve()

    # install ballet
    ballet_pkg = next(p for p in working_set if p.project_name == 'ballet')
    if ballet_pkg is None:
        raise RuntimeError(
            'something is not right; ballet should be installed when running'
            'tests!')

    virtualenv.run(f'python -m pip install -e {ballet_pkg.location}')

    # patch setup.py to unpin dependency
    with d.joinpath('setup.py').open('r') as f:
        setup_contents = f.read()
    setup_contents = re.sub(
        r'ballet==[0-9.+-]+',
        'ballet',
        setup_contents,
    )
    with d.joinpath('setup.py').open('w') as f:
        f.write(setup_contents)

    # install the quickstart project
    virtualenv.run(f'python -m pip install -e {d}')

    # should be installed now, or will fail with error code != 0
    virtualenv.run(
        f'python -c "import {quickstart.package_slug}"',
        capture=True)

    # GH-71 CLI should be enabled and should have correct name
    output = virtualenv.run(
        f'python -m {quickstart.package_slug} --help',
        capture=True
    )
    assert quickstart.package_slug in output



#!/usr/bin/env python

import collections
import json
import pathlib

import git

from ballet.util.log import logger
from ballet.update import TEMPLATE_BRANCH


def create_git_repo():
    repo = git.Repo.init('.')
    with repo.config_writer() as writer:
        writer.set_value('user', 'name', '{{ cookiecutter.full_name }}')
        writer.set_value('user', 'email', '{{ cookiecutter.email }}')
        writer.set_value('github', 'user', '{{ cookiecutter.github_owner }}')
        writer.release()
    repo.git.add('.')
    repo.index.commit('Automatically generated files from ballet quickstart')
    repo.create_remote('origin',
                       'https://github.com/{{ cookiecutter.github_owner }}'
                       '/{{ cookiecutter.project_slug }}')
    repo.create_head(TEMPLATE_BRANCH)


def clean_cookiecutter_context():
    fn = '.cookiecutter_context.json'
    with open(fn, 'r') as f:
        context = json.load(f, object_pairs_hook=collections.OrderedDict)

    # strip _template key
    context['cookiecutter'].pop('_template')

    with open(fn, 'w') as f:
        json.dump(context, f)


def echo():
    fn = pathlib.Path.cwd().resolve()
    logger.info(f'New project created in {fn!s}')


def main():
    clean_cookiecutter_context()
    create_git_repo()
    echo()


if __name__ == '__main__':
    main()



import os
from typing import Dict, Iterable, Optional

import git

from ballet.exc import UnexpectedTravisEnvironmentError
from ballet.util import falsy, truthy
from ballet.util.git import (
    PullRequestBuildDiffer, get_diff_endpoints_from_commit_range,)
from ballet.util.log import logger


def in_travis() -> bool:
    """Check if we are in Travis right now"""
    return len(get_travis_env_vars()) > 0


def get_travis_env_or_fail(name: str) -> str:
    if name in os.environ:
        return os.environ[name]
    else:
        # dump_travis_env_vars()
        raise UnexpectedTravisEnvironmentError(
            f'Missing TRAVIS environment variable: {name}')


def ensure_expected_travis_env_vars(names: Iterable[str]):
    if not set(names).issubset(set(os.environ.keys())):
        raise UnexpectedTravisEnvironmentError


def get_travis_env_vars() -> Dict[str, str]:
    return {
        k: v for k, v in os.environ.items()
        if k.startswith('TRAVIS')
    }


def dump_travis_env_vars():
    logger.info(repr(get_travis_env_vars()))


# TODO delete
def get_travis_pr_num() -> Optional[int]:
    """Return the PR number if the job is a pull request, None otherwise

    Returns:
        int

    See also:
        - <https://docs.travis-ci.com/user/environment-variables/#default-environment-variables>
    """  # noqa E501
    try:
        travis_pull_request = get_travis_env_or_fail('TRAVIS_PULL_REQUEST')
        if falsy(travis_pull_request):
            return None
        else:
            try:
                return int(travis_pull_request)
            except ValueError:
                return None
    except UnexpectedTravisEnvironmentError:
        return None


def is_travis_pr() -> bool:
    """Check if the current job is a pull request build"""
    return get_travis_pr_num() is not None


# TODO delete
def get_travis_branch() -> Optional[str]:
    """Get current branch per Travis environment variables

    If travis is building a PR, then TRAVIS_PULL_REQUEST is truthy and the
    name of the branch corresponding to the PR is stored in the
    TRAVIS_PULL_REQUEST_BRANCH environment variable. Else, the name of the
    branch is stored in the TRAVIS_BRANCH environment variable.

    See also:
        - <https://docs.travis-ci.com/user/environment-variables/#default-environment-variables>
    """  # noqa E501
    try:
        travis_pull_request = get_travis_env_or_fail('TRAVIS_PULL_REQUEST')
        if truthy(travis_pull_request):
            travis_pull_request_branch = get_travis_env_or_fail(
                'TRAVIS_PULL_REQUEST_BRANCH')
            return travis_pull_request_branch
        else:
            travis_branch = get_travis_env_or_fail('TRAVIS_BRANCH')
            return travis_branch
    except UnexpectedTravisEnvironmentError:
        return None


def can_use_travis_differ(repo: Optional[git.Repo]) -> bool:
    """Check if the required travis env vars are set for the travis differ"""
    try:
        ensure_expected_travis_env_vars(
            TravisPullRequestBuildDiffer.EXPECTED_TRAVIS_ENV_VARS)
    except UnexpectedTravisEnvironmentError:
        return False
    else:
        return True


class TravisPullRequestBuildDiffer(PullRequestBuildDiffer):
    """Differ from within Travis test environment

    Requires several environment variables set by Travis to work properly:
        - TRAVIS_BUILD_DIR, to detect the git repo
        - TRAVIS_PULL_REQUEST, to determine whether the build is on a pull
            request
        - TRAVIS_COMMIT_RANGE, to determine which commits to diff.

    Args:
        repo: repo object for the project
    """

    EXPECTED_TRAVIS_ENV_VARS = (
        # 'TRAVIS_BRANCH',
        'TRAVIS_BUILD_DIR',
        'TRAVIS_PULL_REQUEST',
        # 'TRAVIS_PULL_REQUEST_BRANCH',
        'TRAVIS_COMMIT_RANGE',
    )

    def __init__(self, repo: git.Repo = None):
        super().__init__(repo)
        if repo is None:
            repo = self._detect_repo()

    def _check_environment(self):
        ensure_expected_travis_env_vars(
            TravisPullRequestBuildDiffer.EXPECTED_TRAVIS_ENV_VARS)

    def _get_diff_endpoints(self):
        commit_range = get_travis_env_or_fail('TRAVIS_COMMIT_RANGE')
        return get_diff_endpoints_from_commit_range(self.repo, commit_range)

    def _detect_repo(self) -> git.Repo:
        build_dir = get_travis_env_or_fail('TRAVIS_BUILD_DIR')
        return git.Repo(build_dir)



import numpy as np
import pandas as pd
import pytest
import sklearn.base
from sklearn.impute import SimpleImputer

import ballet.eng
import ballet.exc
from ballet.util.testing import (
    assert_array_equal, assert_frame_equal, assert_series_equal,
    assert_series_not_equal,)


def test_no_fit_mixin():
    class _Foo(ballet.eng.NoFitMixin):
        pass
    a = _Foo()

    assert hasattr(a, 'fit')

    # method should exist
    a.fit('X')
    a.fit('X', y=None)


def test_base_transformer():
    a = ballet.eng.BaseTransformer()

    assert isinstance(a, sklearn.base.BaseEstimator)
    assert hasattr(a, 'fit')


def test_simple_function_transformer():
    def func(x): return x + 5
    data = np.arange(30)

    trans = ballet.eng.SimpleFunctionTransformer(func)
    trans.fit(data)
    data_trans = trans.transform(data)
    data_func = func(data)

    assert_array_equal(data_trans, data_func)


def test_simple_function_transformer_str_repr():
    trans = ballet.eng.SimpleFunctionTransformer(lambda x: x)
    for func in [str, repr]:
        s = func(trans)
        assert len(s) > 0


def test_grouped_function_transformer():
    df = pd.DataFrame(
        data={
            'country': ['USA', 'USA', 'USA', 'Canada', 'Fiji'],
            'year': [2001, 2002, 2003, 2001, 2001],
            'length': [1, 2, 3, 4, 5],
            'width': [1.0, 1.0, 7.5, 9.0, 11.0],
        }
    ).set_index(['country', 'year']).sort_index()

    # with groupby kwargs, produces a df
    func = np.sum
    trans = ballet.eng.GroupedFunctionTransformer(
        func, groupby_kwargs={'level': 'country'})
    trans.fit(df)
    result = trans.transform(df)
    expected_result = df.groupby(level='country').apply(func)
    assert_frame_equal(result, expected_result)

    # without groupby kwargs, produces a series
    func = np.min
    trans = ballet.eng.GroupedFunctionTransformer(func)
    trans.fit(df)
    result = trans.transform(df)
    expected_result = df.pipe(func)
    assert_series_equal(result, expected_result)


@pytest.fixture
def sample_data():
    X_tr = pd.DataFrame(
        data={
            'name': ['A', 'A', 'A', 'B', 'B', 'C', 'C'],
            'year': [2001, 2002, 2003, 2001, 2002, 2001, 2003],
            'value': [1, 2, np.nan, 4, 4, 5, np.nan],
            'size': [3, 5, 5, 5, 5, np.nan, 4],
        }
    ).set_index(['name', 'year']).sort_index()

    X_te = pd.DataFrame(
        data={
            'name': ['A', 'B', 'C'],
            'year': [2004, 2004, 2004],
            'value': [np.nan, 1.5, np.nan],
            'size': [4, 1, np.nan],
        }
    ).set_index(['name', 'year']).sort_index()

    return X_tr, X_te


@pytest.fixture
def groupby_kwargs():
    return {'level': 'name'}


@pytest.fixture
def individual_transformer():
    return SimpleImputer()


@pytest.fixture
def groupwise_transformer(groupby_kwargs, individual_transformer):
    # mean-impute within groups
    return ballet.eng.GroupwiseTransformer(
        individual_transformer,
        groupby_kwargs=groupby_kwargs,
        column_selection=['value'],
    )


def test_groupwise_transformer_can_fit(sample_data, groupwise_transformer):
    X_tr, X_te = sample_data
    groupwise_transformer.fit(X_tr)


def test_groupwise_transformer_can_transform(
    sample_data, groupwise_transformer
):
    X_tr, X_te = sample_data
    groupwise_transformer.fit(X_tr)

    result_tr = groupwise_transformer.transform(X_tr)
    expected_tr = X_tr.copy()
    expected_tr['value'] = np.array([1, 2, 1.5, 4, 4, 5, 5])
    expected_tr = expected_tr.drop('size', axis=1)
    assert_frame_equal(result_tr, expected_tr)

    result_te = groupwise_transformer.transform(X_te)
    expected_te = X_te.copy()
    expected_te['value'] = np.array([1.5, 1.5, 5])
    expected_te = expected_te.drop('size', axis=1)
    assert_frame_equal(result_te, expected_te)


def test_groupwise_transformer_raise_on_new_group(
    sample_data, individual_transformer, groupby_kwargs
):
    X_tr, X_te = sample_data

    # mean-impute within groups
    groupwise_transformer = ballet.eng.GroupwiseTransformer(
        individual_transformer,
        groupby_kwargs=groupby_kwargs,
        handle_unknown='error',
    )

    groupwise_transformer.fit(X_tr)

    X_te = X_te.copy().reset_index()
    X_te.loc[0, 'name'] = 'Z'  # new group
    X_te = X_te.set_index(['name', 'year'])

    with pytest.raises(ballet.exc.BalletError):
        groupwise_transformer.transform(X_te)


def test_groupwise_transformer_ignore_on_new_group(
    sample_data, individual_transformer, groupby_kwargs
):

    X_tr, X_te = sample_data

    groupwise_transformer = ballet.eng.GroupwiseTransformer(
        individual_transformer,
        groupby_kwargs=groupby_kwargs,
        handle_unknown='ignore',
    )

    groupwise_transformer.fit(X_tr)

    X_te = X_te.copy().reset_index()
    X_te.loc[0, 'name'] = 'Z'  # new group
    X_te = X_te.set_index(['name', 'year'])

    result = groupwise_transformer.transform(X_te)

    # the first group, Z, is new, and values are passed through, so such
    # be nan
    expected = X_te.copy()
    expected['value'] = np.array([np.nan, 1.5, 5.0])
    expected['size'] = np.array([4.0, 1.0, 4.0])

    assert_frame_equal(result, expected)


def test_groupwise_transformer_raise_on_transform_error(
    sample_data, groupby_kwargs
):
    X_tr, X_te = sample_data

    exc = Exception

    class TransformErrorTransformer(ballet.eng.BaseTransformer):
        def transform(self, X, **transform_kwargs):
            raise exc

    groupwise_transformer = ballet.eng.GroupwiseTransformer(
        TransformErrorTransformer(),
        groupby_kwargs=groupby_kwargs,
        handle_error='error',
    )

    groupwise_transformer.fit(X_tr)

    with pytest.raises(exc):
        groupwise_transformer.transform(X_tr)


def test_groupwise_transformer_ignore_on_transform_error(sample_data):
    X_tr, X_te = sample_data

    exc = Exception

    class TransformErrorTransformer(ballet.eng.BaseTransformer):
        def transform(self, X, **transform_kwargs):
            raise exc

    groupwise_transformer = ballet.eng.GroupwiseTransformer(
        TransformErrorTransformer(),
        groupby_kwargs={'level': 'name'},
        handle_error='ignore',
    )

    groupwise_transformer.fit(X_tr)

    result_tr = groupwise_transformer.transform(X_tr)
    expected_tr = X_tr

    assert_frame_equal(result_tr, expected_tr)

    result_te = groupwise_transformer.transform(X_te)
    expected_te = X_te
    assert_frame_equal(result_te, expected_te)


def test_conditional_transformer_both_satisfied(sample_data):
    X_tr, X_te = sample_data

    t = ballet.eng.ConditionalTransformer(
        lambda ser: ser.sum() > 0,
        lambda ser: ser + 1,
    )

    # all the features are selected by sum > 0
    t.fit(X_tr)
    result_tr = t.transform(X_tr)
    for col in ['value', 'size']:
        assert_series_not_equal(result_tr[col], X_tr[col])

    result_te = t.transform(X_te)
    for col in ['value', 'size']:
        assert_series_not_equal(result_te[col], X_te[col])


def test_conditional_transformer_one_satisfied(sample_data):
    X_tr, X_te = sample_data

    t = ballet.eng.ConditionalTransformer(
        lambda ser: (ser.dropna() >= 3).all(),
        lambda ser: ser.fillna(0) + 1,
    )

    t.fit(X_tr)
    result_tr = t.transform(X_tr)
    result_te = t.transform(X_te)

    # only 'size' is selected by the condition
    assert_series_not_equal(result_tr['size'], X_tr['size'])
    assert_series_not_equal(result_te['size'], X_te['size'])

    # 'value' is not selected by the condition, has items less than 3
    assert_series_equal(result_tr['value'], X_tr['value'])
    assert_series_equal(result_te['value'], X_te['value'])


def test_conditional_transformer_unsatisfy_transform(sample_data):
    X_tr, X_te = sample_data

    t = ballet.eng.ConditionalTransformer(
        lambda ser: (ser.dropna() >= 3).all(),
        lambda ser: ser,
        lambda ser: ser.fillna(0) - 1,
    )

    t.fit(X_tr)
    result_tr = t.transform(X_tr)
    result_te = t.transform(X_te)

    # size is transformed by satisfy condition, but passed through
    assert_series_equal(result_tr['size'], X_tr['size'])
    assert_series_equal(result_te['size'], X_te['size'])

    # value is transformed by unsatisfy condition and is not equal
    assert_series_not_equal(result_tr['value'], X_tr['value'])
    assert_series_not_equal(result_te['value'], X_te['value'])


def test_subset_transformer_identity(sample_data):
    """After passing through a column unchanged, the entire df is the same as before"""  # noqa
    X_tr, X_te = sample_data

    t = ballet.eng.SubsetTransformer('value', None)
    result_tr = t.fit_transform(X_tr)
    result_te = t.transform(X_te)

    assert_frame_equal(result_tr, X_tr)
    assert_frame_equal(result_te, X_te)


def test_subset_transformer_mutate(sample_data):
    """After modifying one column, that column is different and the complement is the same"""  # noqa
    X_tr, X_te = sample_data

    input = 'size'
    t = ballet.eng.SubsetTransformer(input, lambda x: x + 1)
    result_tr = t.fit_transform(X_tr)
    result_te = t.transform(X_te)

    # the input col is modified
    assert_series_not_equal(result_tr[input], X_tr[input])
    assert_series_not_equal(result_te[input], X_te[input])

    # the complement is passed through unchanged
    complement = [col for col in X_tr.columns if col != input]
    assert_frame_equal(result_tr[complement], X_tr[complement])
    assert_frame_equal(result_te[complement], X_te[complement])



from inspect import signature
from typing import Optional, Tuple

from funcy import cached_property

import ballet.pipeline
from ballet.transformer import RobustTransformer, make_robust_transformer
from ballet.util.typing import OneOrMore, TransformerLike

__all__ = ('Feature', )


class Feature:
    """A feature definition

    Conceptually, a feature definition is a learned function that maps raw
    variables in one data instance to a vector of feature values. A feature
    definition can produce either a scalar feature value for each instance
    or a vector of feature values, as in the case of an embedding technique
    like PCA or the one-hot encoding of a categorical variable.

    Args:
        input: required columns from the input dataframe needed for the
            transformation
        transformer: transformer, sequence of transformers, or ``None``. A
            "transformer" is an instance of a class that provides a
            fit/transform-style learned transformation. Alternately, a
            callable can be provided, either by itself or in a list, in
            which case it will be converted into a
            :py:class:``FunctionTransformer`` for convenience. If ``None``
            is provided, it will be replaced with the
            :py:class:``IdentityTransformer``.
        name: name of the feature
        description: description of the feature
        output: ordered sequence of names of features
            produced by this transformer
        source: the module in which this feature was defined
        options: options
    """

    def __init__(
        self,
        input: OneOrMore[str],
        transformer: OneOrMore[TransformerLike],
        name: Optional[str] = None,
        description: Optional[str] = None,
        output: OneOrMore[str] = None,
        source: Optional[str] = None,
        options: dict = None
    ):
        self.input = input
        self.transformer = make_robust_transformer(transformer)
        self.name = name
        self.description = description
        self.output = output  # unused
        self.source = source
        self.options = options if options is not None else {}

    @cached_property
    def _init_attr_list(self):
        return list(signature(self.__init__).parameters)

    def __repr__(self):
        indent = 4
        attr_list = self._init_attr_list
        attrs_str = ',\n'.join(
            '{indent}{attr_name}={attr_val!s}'.format(
                indent=' ' * indent,
                attr_name=attr,
                attr_val=getattr(self, attr)
            ) for attr in attr_list
        )
        return '{clsname}(\n{attrs_str}\n)'.format(
            clsname=type(self).__name__, attrs_str=attrs_str)

    def as_input_transformer_tuple(
        self
    ) -> Tuple[OneOrMore[str], RobustTransformer, dict]:
        """Return an tuple for passing to DataFrameMapper constructor"""
        return self.input, self.transformer, {'alias': self.output}

    def as_feature_engineering_pipeline(
        self
    ) -> ballet.pipeline.FeatureEngineeringPipeline:
        """Return standalone FeatureEngineeringPipeline with this feature"""
        return ballet.pipeline.FeatureEngineeringPipeline(self)

    @property
    def author(self) -> Optional[str]:
        """The author of this feature if it can be inferred from its source

        The author can be inferred if the module the feature was defined in
        follows the pattern
        ``package.subpackage.user_username.feature_featurename``. Otherwise,
        returns ``None``.
        """
        if self.source:
            pieces = self.source.rsplit('.', maxsplit=2)
            if len(pieces) >= 2:
                user_str = pieces[-2]
                if user_str.startswith('user_'):
                    return user_str[len('user_'):]

        return None

    _pipeline = None

    @property
    def pipeline(self) -> ballet.pipeline.FeatureEngineeringPipeline:
        """A feature engineering pipeline containing just this feature"""
        if self._pipeline is None:
            self._pipeline = self.as_feature_engineering_pipeline()

        return self._pipeline

    def fit(self, X, y=None):
        """Fit feature.pipeline"""
        return self.pipeline.fit(X, y=y)

    def transform(self, X):
        """Transform data using feature.pipeline"""
        return self.pipeline.transform(X)

    def fit_transform(self, X, y=None):
        """Fit feature.pipeline and then transform data"""
        return self.fit(X, y=y).transform(X)



from unittest.mock import patch

import pytest

from ballet.util.ci import (
    TravisPullRequestBuildDiffer, get_travis_branch, is_travis_pr,)
from ballet.util.git import make_commit_range
from tests.util import make_mock_commit, make_mock_commits


@pytest.fixture
def commit_range():
    return 'HEAD^..HEAD'


@pytest.mark.parametrize(
    ('key, value, expected'),
    [
        ('TRAVIS_PULL_REQUEST', '7', True),
        ('TRAVIS_PULL_REQUEST', 'true', False),
        ('TRAVIS_PULL_REQUEST', 'FALSE', False),
        ('TRAVIS_PULL_REQUEST', 'false', False),
        ('TRAVIS_PULL_REQUEST', 'abcd', False),
        ('UNRELATED', '', False),
    ],
)
def test_is_travis_pr(key, value, expected):
    with patch.dict('os.environ', {key: value}, clear=True):
        actual = is_travis_pr()
        assert actual == expected


def test_get_travis_branch():
    # matrix of env dict, expected result
    matrix = (
        ({
            'TRAVIS_PULL_REQUEST': 'false',
            'TRAVIS_PULL_REQUEST_BRANCH': '',
            'TRAVIS_BRANCH': 'master',
        }, 'master'),
        ({
            'TRAVIS_PULL_REQUEST': 'false',
            'TRAVIS_PULL_REQUEST_BRANCH': '',
            'TRAVIS_BRANCH': 'foo',
        }, 'foo'),
        ({
            'TRAVIS_PULL_REQUEST': '1',
            'TRAVIS_PULL_REQUEST_BRANCH': 'foo',
            'TRAVIS_BRANCH': 'master',
        }, 'foo'),
        ({}, None),
    )

    for env, expected in matrix:
        with patch.dict('os.environ', env, clear=True):
            actual = get_travis_branch()
            assert actual == expected


def test_travis_pull_request_build_differ(mock_repo, commit_range):
    repo = mock_repo
    make_mock_commits(repo, n=3)

    travis_env_vars = {
        'TRAVIS_BUILD_DIR': repo.working_tree_dir,
        'TRAVIS_PULL_REQUEST': '1',
        'TRAVIS_COMMIT_RANGE': commit_range,
    }
    with patch.dict('os.environ', travis_env_vars, clear=True):
        differ = TravisPullRequestBuildDiffer(repo)
        expected_a = repo.rev_parse('HEAD^')
        expected_b = repo.rev_parse('HEAD')
        actual_a, actual_b = differ._get_diff_endpoints()
        assert actual_a == expected_a
        assert actual_b == expected_b


def test_travis_pull_request_build_differ_on_mock_commits(mock_repo):
    repo = mock_repo
    n = 4
    i = 0
    feature_branch_name = 'alice/feature-x'

    make_mock_commit(repo, path='readme.txt')
    expected_merge_base = repo.head.commit
    feature_branch = repo.create_head(feature_branch_name)

    # make commits on branch master
    commits = make_mock_commits(repo, n=3, filename='blah{i}.txt')
    master = repo.heads.master

    # make commits on feature branch
    feature_branch.checkout()
    commits = make_mock_commits(repo, n=n)
    end_commit = commits[-1]

    commit_range = make_commit_range(
        master, end_commit)

    travis_env_vars = {
        'TRAVIS_BUILD_DIR': repo.working_tree_dir,
        'TRAVIS_PULL_REQUEST': '1',
        'TRAVIS_COMMIT_RANGE': commit_range,
    }
    with patch.dict('os.environ', travis_env_vars, clear=True):
        differ = TravisPullRequestBuildDiffer(repo)
        a, b = differ._get_diff_endpoints()
        assert a == expected_merge_base
        assert b == end_commit

        diffs = differ.diff()

        # there should be n diff objects, they should show files
        # 0 to n-1. merge base just created readme.txt, so all
        # files on feature branch are new.
        assert len(diffs) == n
        j = i
        for diff in diffs:
            assert diff.change_type == 'A'
            assert diff.b_path == f'file{j}.py'
            j += 1



from unittest.mock import ANY, patch

import funcy
import pytest
from click.testing import CliRunner

import ballet
from ballet.cli import cli as _cli


@pytest.fixture
def cli():
    runner = CliRunner()
    return funcy.partial(runner.invoke, _cli)


def test_cli_version(cli):
    result = cli('--version')
    assert result.exit_code == 0
    assert ballet.__version__ in result.output


@pytest.mark.parametrize(
    'create_github_repo',
    [True, False],
    ids=['--create-github-repo', '--no-create-github-repo'],
)
@patch('ballet.templating.render_project_template')
def test_quickstart(mock_render, cli, create_github_repo):
    cmd = 'quickstart'
    if create_github_repo:
        cmd += ' --create-github-repo'
    result = cli(cmd)

    # want to do a partial match on kwargs
    # compat: call.kwargs introduced in py3.8
    mock_render.assert_called_once()
    _, kwargs = mock_render.call_args
    assert kwargs['create_github_repo'] == create_github_repo

    assert 'Generating new ballet project' in result.output


@patch('ballet.update.update_project_template')
def test_update_project_template(mock_update, cli):
    result = cli('update-project-template --push')
    mock_update.assert_called_once_with(
        push=True, project_template_path=None)
    assert 'Updating project template' in result.output


@patch('ballet.templating.start_new_feature')
def test_start_new_feature(mock_start, cli):
    result = cli('start-new-feature')
    mock_start.assert_called_once_with(branching=True)
    assert 'Starting new feature' in result.output


@patch('ballet.project.Project.from_path')
@patch('ballet.validation.main.validate')
def test_validate_all(mock_validate, mock_from_path, cli):
    result = cli('validate -A')
    mock_validate.assert_called_once_with(ANY, True, True, True, True)
    assert 'Validating project' in result.output



import numpy as np
import pandas as pd
import pytest

from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline


@pytest.fixture(
    params=[
        IdentityTransformer(),
        [IdentityTransformer()],
        [None, IdentityTransformer(), lambda x: x],
        Feature(['foo', 'bar'], IdentityTransformer()),
        [None, IdentityTransformer(),
            Feature(['foo', 'bar'], IdentityTransformer())],
    ],
    ids=[
        'scalar',
        'list of transformer',
        'list of mixed',
        'nested feature',
        'list of mixed and nested features',
    ]
)
def inputs(request):
    input = ['foo', 'bar']
    transformer = request.param
    return input, transformer


def test_init(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    assert isinstance(mapper, FeatureEngineeringPipeline)


def test_fit(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    df = pd.util.testing.makeCustomDataframe(5, 2)
    df.columns = ['foo', 'bar']
    mapper.fit(df)


def test_transform(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    mapper = FeatureEngineeringPipeline(feature)
    df = pd.util.testing.makeCustomDataframe(5, 2)
    df.columns = ['foo', 'bar']
    mapper.fit(df)
    X = mapper.transform(df)
    assert np.shape(X) == (5, len(inputs))



import inspect
from typing import Callable, List, Tuple, Union

import pandas as pd
from funcy import decorator, func_partial, ignore
from funcy.decorators import Call
from stacklog import stacklog

from ballet.exc import (
    FeatureRejected, InvalidFeatureApi, InvalidProjectStructure,
    NoFeaturesCollectedError, SkippedValidationTest,)
from ballet.feature import Feature
from ballet.project import Project
from ballet.util.log import logger
from ballet.validation.common import (
    get_accepted_features, get_proposed_feature, load_spec,)

# helpful for log parsing
PRUNER_MESSAGE = 'Found Redundant Feature: '


@decorator
def validation_stage(call: Call, message: str):
    call = stacklog(logger.info,
                    f'Ballet Validation: {message}',
                    conditions=[(SkippedValidationTest, 'SKIPPED')])(call)
    call = ignore(SkippedValidationTest)(call)
    return call()


def _load_validator_class_params(
    project: Project, config_key: str
) -> Callable:
    """Load validator class according to config_key with optional params

    At the provided key, the config should show an entry in one of two forms:
    1. The fully-qualified class name of the validator (str)
    2. A yaml hash with the key `name` mapping to the fully-qualified class
    name of the validator, and optionally, the key `params` mapping to hash
    of keyword arguments to be passed to the validator class.

    If `params` is provided, then they are partially applied to the
    validator class ``__init__`` method such that calls to create an
    instance of the validator class have the given params set as keyword
    arguments.

    For example, if the yaml file looks like::

       foo:
         bar:
           validation:
               feature_accepter:
                 name: baz.qux.MyFeatureAccepter
                 params:
                   key1: value1

    Then::

       make_validator = _load_validator_class_params(project, 'foo.bar.validation.feature_accepter')

    would result in the following equivalence::

       make_validator(arg)
       baz.qux.MyFeatureAccepter(arg, key1=value1)
    """  # noqa E501
    spec = project.config.get(config_key)
    cls, params = load_spec(spec)
    return func_partial(cls, **params)


def _load_validation_data(
    project: Project
) -> Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]:
    """Load the validation data split

    The validation data split should be given by the key `validation.split` in
    the project's config. If the key is not present, or the `load_data` method
    does not support loading splits, then the default dataset is returned.
    """
    kwargs = {}
    try:
        val_split = project.config.validation.split
        sig = inspect.signature(project.api.load_data)
        if 'split' in sig.parameters:
            kwargs['split'] = val_split
    except Exception:
        pass
    X_df, y_df = project.api.load_data(**kwargs)
    return X_df, y_df


@validation_stage('checking project structure')
def _check_project_structure(project: Project, force: bool = False):
    if not force and project.on_master:
        raise SkippedValidationTest('Not on feature branch')

    validator_class = _load_validator_class_params(
        project, 'validation.project_structure_validator')
    validator = validator_class(project)
    result = validator.validate()
    if not result:
        raise InvalidProjectStructure


@validation_stage('validating feature API')
def _validate_feature_api(project: Project, force: bool = False):
    """Validate feature API"""
    if not force and project.on_master:
        raise SkippedValidationTest('Not on feature branch')

    validator_class = _load_validator_class_params(
        project, 'validation.feature_api_validator')
    validator = validator_class(project)
    result = validator.validate()
    if not result:
        raise InvalidFeatureApi


@validation_stage('evaluating feature performance')
def _evaluate_feature_performance(project: Project, force: bool = False):
    """Evaluate feature performance"""
    if not force and project.on_master:
        raise SkippedValidationTest('Not on feature branch')

    X_df, y_df = project.api.load_data()
    X_df_val, y_df_val = _load_validation_data(project)

    encoder = project.api.encoder
    y_val = encoder.fit(y_df).transform(y_df_val)

    features = project.api.features
    proposed_feature = get_proposed_feature(project)
    accepted_features = get_accepted_features(features, proposed_feature)

    accepter_class = _load_validator_class_params(
        project, 'validation.feature_accepter')
    accepter = accepter_class(
        X_df, y_df, X_df_val, y_val, accepted_features, proposed_feature)
    accepted = accepter.judge()

    if not accepted:
        raise FeatureRejected


@validation_stage('pruning existing features')
def _prune_existing_features(
    project: Project, force: bool = False
) -> List[Feature]:
    """Prune existing features"""
    if not force and not project.on_master:
        raise SkippedValidationTest('Not on master')

    try:
        # if on master but not after merge, then we diff master with itself
        # and collect no features.
        proposed_feature = get_proposed_feature(project)
    except NoFeaturesCollectedError:
        raise SkippedValidationTest('No features collected')

    X_df, y_df = project.api.load_data()
    X_df_val, y_df_val = _load_validation_data(project)

    encoder = project.api.encoder
    y_val = encoder.fit(y_df).transform(y_df_val)

    features = project.api.features
    accepted_features = get_accepted_features(features, proposed_feature)

    pruner_class = _load_validator_class_params(
        project, 'validation.feature_pruner')
    pruner = pruner_class(
        X_df, y_df, X_df_val, y_val, accepted_features, proposed_feature)
    redundant_features = pruner.prune()

    # "propose removal"
    for feature in redundant_features:
        logger.info(PRUNER_MESSAGE + feature.source)

    return redundant_features


def validate(project: Project,
             check_project_structure: bool,
             check_feature_api: bool,
             evaluate_feature_acceptance: bool,
             evaluate_feature_pruning: bool):
    """Entrypoint for 'ballet validate' command in ballet projects"""
    if check_project_structure:
        _check_project_structure(project)
    if check_feature_api:
        _validate_feature_api(project)
    if evaluate_feature_acceptance:
        _evaluate_feature_performance(project)
    if evaluate_feature_pruning:
        _prune_existing_features(project)



import numpy as np
import pandas as pd
import pandas.util.testing as pdt
import pytest

from ballet.util.testing import (
    assert_array_equal, assert_array_not_equal, assert_frame_equal,
    assert_frame_not_equal, assert_index_equal, assert_index_not_equal,
    assert_series_equal, assert_series_not_equal,)


def test_assert_array_equal():
    a = np.arange(10 * 7).reshape(10, 7)
    b = a.copy()
    assert_array_equal(a, b)

    c = np.arange(9).reshape(3, 3)
    d = c + 1

    with pytest.raises(AssertionError):
        assert_array_equal(c, d)


def test_assert_array_not_equal():
    a = np.arange(10 * 7).reshape(10, 7)
    b = a.copy()

    with pytest.raises(AssertionError):
        assert_array_not_equal(a, b)

    c = np.arange(9).reshape(3, 3)
    d = c + 1

    assert_array_not_equal(c, d)


def test_assert_frame_equal():
    a = pdt.makeCustomDataframe(
        10, 7, data_gen_f=lambda row, col: row * col)
    b = a.copy()
    assert_frame_equal(a, b)

    c = a + 1
    with pytest.raises(AssertionError):
        assert_frame_equal(a, c)

    d = pdt.makeCustomDataframe(11, 9)
    with pytest.raises(AssertionError):
        assert_frame_equal(a, d)

    e = pdt.makeTimeSeries()
    with pytest.raises(AssertionError):
        assert_frame_equal(a, e)

    f = pd.DataFrame([1, 2, 3, 4])
    g = pd.Series([1, 2, 3, 4])
    with pytest.raises(AssertionError):
        assert_frame_equal(f, g)


def test_assert_frame_not_equal():
    a = pdt.makeCustomDataframe(
        10, 7, data_gen_f=lambda row, col: row * col)
    b = a.copy()
    with pytest.raises(AssertionError):
        assert_frame_not_equal(a, b)

    c = a + 1
    assert_frame_not_equal(a, c)

    d = pdt.makeCustomDataframe(11, 9)
    assert_frame_not_equal(a, d)

    e = pdt.makeTimeSeries()
    assert_frame_not_equal(a, e)

    f = pd.DataFrame([1, 2, 3, 4])
    g = pd.Series([1, 2, 3, 4])
    assert_frame_not_equal(f, g)


def test_assert_series_equal():
    a = pd.Series(np.arange(21))
    b = a.copy()
    assert_series_equal(a, b)

    c = a + 1
    with pytest.raises(AssertionError):
        assert_series_equal(a, c)

    d = pd.Series(np.arange(17))
    with pytest.raises(AssertionError):
        assert_series_equal(a, d)

    e = pdt.makeDataFrame()
    with pytest.raises(AssertionError):
        assert_series_equal(a, e)


def test_assert_series_not_equal():
    a = pd.Series(np.arange(21))
    b = a.copy()
    with pytest.raises(AssertionError):
        assert_series_not_equal(a, b)

    c = a + 1
    assert_series_not_equal(a, c)

    d = pd.Series(np.arange(17))
    assert_series_not_equal(a, d)

    e = pdt.makeDataFrame()
    assert_series_not_equal(a, e)


def test_assert_index_equal():
    a = pd.Index(np.arange(21))
    b = a.copy()
    assert_index_equal(a, b)

    c = a + 1
    with pytest.raises(AssertionError):
        assert_index_equal(a, c)

    d = pd.Index(np.arange(17))
    with pytest.raises(AssertionError):
        assert_index_equal(a, d)

    e = pdt.makeDataFrame()
    with pytest.raises(AssertionError):
        assert_index_equal(a, e)


def test_assert_index_not_equal():
    a = pd.Index(np.arange(21))
    b = a.copy()
    with pytest.raises(AssertionError):
        assert_index_not_equal(a, b)

    c = a + 1
    assert_index_not_equal(a, c)

    d = pd.Index(np.arange(17))
    assert_index_not_equal(a, d)

    e = pdt.makeDataFrame()
    assert_index_not_equal(a, e)



import numpy as np
import pytest

import ballet.discovery
from ballet.discovery import countunique, discover
from ballet.eng import NullFiller
from ballet.feature import Feature
from ballet.util.testing import assert_array_equal
from tests.util import FragileTransformer


@pytest.mark.parametrize(
    'z,expected',
    [
        (np.array([[1, 2, 3]]).T, np.array([3])),
        (np.array([[1, 2, 3], [2, 2, 3]]).T, np.array([3, 2])),
    ]
)
def test_countunique(z, expected):
    nunique = countunique(z)
    assert_array_equal(nunique, expected)


def test_discover(sample_data):
    features = [
        Feature(
            'size', NullFiller(0),
            source='foo.features.contrib.user_a.feature_1'
        ),
        Feature(
            'strength', NullFiller(100),
            source='foo.features.contrib.user_b.feature_1'
        )
    ]
    X_df, y_df = sample_data.X, sample_data.y
    y = np.asfarray(y_df)

    df = discover(features, X_df, y_df, y)

    expected_cols = {
        'name', 'description', 'input', 'transformer', 'primitives', 'output',
        'author', 'source', 'mutual_information',
        'conditional_mutual_information', 'mean', 'std', 'variance', 'nunique',
    }
    actual_cols = df.columns
    assert not expected_cols.symmetric_difference(actual_cols)

    assert df.shape[0] == len(features)

    # test filter
    input = 'size'
    df = discover(features, X_df, y_df, y, input=input)
    assert df.shape[0] == len([
        feature
        for feature in features
        if feature.input == input or input in feature.input
    ])

    # test no data available
    # have to clear cache, as values on data already known
    ballet.discovery._summarize_feature.memory.clear()
    df = discover(features, None, None, None)
    assert df.shape[0] == len(features)
    actual_cols = df.columns
    assert not expected_cols.symmetric_difference(actual_cols)
    assert np.isnan(df['mean'].at[0])


def test_discover_feature_error(sample_data):
    features = [
        Feature('size', FragileTransformer()),
    ]
    X_df, y_df = sample_data.X, sample_data.y
    y = np.asfarray(y_df)

    df = discover(features, X_df, y_df, y)

    assert df.shape[0] == len(features)
    assert np.isnan(df['mean'].at[0])



from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from ballet.feature import Feature
from ballet.util import asarray2d
from ballet.validation.base import FeaturePerformanceEvaluator
from ballet.validation.entropy import estimate_entropy

LAMBDA_1_ADJUSTMENT = 64
LAMBDA_2_ADJUSTMENT = 64


def _concat_datasets(
    feature_df_map: Dict[Feature, pd.DataFrame],
    n_samples: int = 0,
    omit: Optional[List[Feature]] = None
) -> np.ndarray:
    if omit is None:
        omit = []

    filtered_dfs = [
        np.array(feature_df_map[feature])
        for feature in feature_df_map
        if feature not in omit
    ]

    if not filtered_dfs:
        return np.zeros((n_samples, 1))

    return asarray2d(np.concatenate(filtered_dfs, axis=1))


def _compute_lmbdas(
    unnorm_lmbda_1: float,
    unnorm_lmbda_2: float,
    features_by_src: Dict[str, np.ndarray],
) -> Tuple[float, float]:
    num_features = len(features_by_src)
    num_feature_cols = sum(
        features_by_src[feat_src].shape[1]
        for feat_src in features_by_src
    )
    # if there are no features, then don't adjust the lambdas
    # num_feature_cols = max(1, num_feature_cols)
    lmbda_1 = unnorm_lmbda_1 / num_features
    lmbda_2 = unnorm_lmbda_2 / num_feature_cols
    return lmbda_1, lmbda_2


def _compute_threshold(
    lmbda_1: float,
    lmbda_2: float,
    n_feature_cols: int,
    n_omitted_cols: int = 0
) -> float:
    return lmbda_1 + lmbda_2 * (n_feature_cols - n_omitted_cols)


@dataclass
class GFSSFIterationInfo:
    i: int
    n_samples: int
    candidate_feature: Feature
    candidate_cols: int
    candidate_cmi: float
    omitted_feature: Feature
    omitted_cols: int
    omitted_cmi: float
    statistic: float
    threshold: float
    delta: float

    def __str__(self):
        def format(v):
            if isinstance(v, (float, np.float_)):
                return f'{v:.4e}'
            elif isinstance(v, Feature):
                return v.source or '<live object>'
            return str(v)
        return ', '.join(
            f'{k}={format(v)}'
            for k, v in self.__dict__.items()
        )


class GFSSFPerformanceEvaluator(FeaturePerformanceEvaluator):
    """A feature performance evaluator that uses a modified version of GFSSF[1]

    Attributes:
        lmbda_1: GFSSF parameter used to calculate the information
            threshold. Default is a function of the entropy of y.
        lmbda_2: GFSSF parameter used to calculate the information
            threshold. Default is a function of the entropy of y.
        lambda_1_adjustment: Adjustment to estimated entropy used to
            calculate lmbda_1.
        lambda_2_adjustment: Adjustment to estimated entropy used to
            calculate lmbda_2.

    References:
        [1] H. Li, X. Wu, Z. Li and W. Ding, "Group Feature Selection
            with Streaming Features," 2013 IEEE 13th International
            Conference on Data Mining, Dallas, TX, 2013, pp. 1109-1114.
            doi: 10.1109/ICDM.2013.137
    """

    def __init__(
        self,
        *args,
        lmbda_1: float = 0.0,
        lmbda_2: float = 0.0,
        lambda_1_adjustment: float = LAMBDA_1_ADJUSTMENT,
        lambda_2_adjustment: float = LAMBDA_2_ADJUSTMENT
    ):
        super().__init__(*args)
        self.y_val = asarray2d(self.y_val)
        if lmbda_1 <= 0:
            lmbda_1 = estimate_entropy(self.y_val) / lambda_1_adjustment
        if lmbda_2 <= 0:
            lmbda_2 = estimate_entropy(self.y_val) / lambda_2_adjustment
        self.lmbda_1 = lmbda_1
        self.lmbda_2 = lmbda_2

    def __str__(self):
        cls = super().__str__()
        return \
            f'{cls}: lmbda_1={self.lmbda_1:0.4f}, lmbda_2={self.lmbda_2:0.4f}'

    def _get_feature_df_map(self):
        all_features = [*self.features, self.candidate_feature]

        def as_features(feature):
            return (
                feature
                .as_feature_engineering_pipeline()
                .fit(self.X_df, y=self.y_df)
                .transform(self.X_df_val)
            )

        # map feature defintion "id" -> feature values
        feature_df_map = {
            feature: as_features(feature)
            for feature in all_features
        }

        return feature_df_map



from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from ballet.eng.base import SimpleFunctionTransformer
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.util import asarray2d
from ballet.validation.feature_pruning.validator import (
    GFSSFPruner, NoOpPruner, RandomPruner,)
from tests.util import load_regression_data


def test_noop_pruner():
    X_df = y_df = y = None
    existing_features = []
    feature = None

    expected = []

    pruner = NoOpPruner(X_df, y_df, X_df, y, existing_features, feature)
    actual = pruner.prune()

    assert expected == actual


@patch('random.choice')
@patch('random.uniform', return_value=0.0)  # makes sure pruning is performed
def test_random_pruner(mock_uniform, mock_choice):
    existing_features = [MagicMock(), MagicMock()]

    mock_choice.return_value = existing_features[0]  # the pruned feature
    expected = [existing_features[0]]

    X_df = y_df = y = None
    feature = None

    pruner = RandomPruner(X_df, y_df, X_df, y, existing_features, feature)
    actual = pruner.prune()

    assert expected == actual


@pytest.fixture
def sample_data():
    X_df, y_df = load_regression_data(
        n_informative=1, n_uninformative=14)
    y = y_df
    return X_df, y_df, y


def test_gfssf_pruner_prune_exact_replicas(sample_data):
    X_df, y_df, y = sample_data

    feature_1 = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    feature_2 = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='2nd Feature')
    gfssf_pruner = GFSSFPruner(
        X_df, y_df, X_df, y, [feature_1], feature_2)

    redunant_features = gfssf_pruner.prune()
    assert feature_1 in redunant_features, \
        'Exact replica features should be pruned'


@pytest.mark.xfail
def test_gfssf_pruner_prune_weak_replicas(sample_data):
    X_df, y_df, y = sample_data

    def add_noise(X):
        X = asarray2d(X)
        return X + np.random.normal(0, 0.5, X.shape)

    noise_transformer = SimpleFunctionTransformer(add_noise)
    feature_weak = Feature(
        input='A_0',
        transformer=noise_transformer,
        source='1st Feature')
    feature_strong = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='2nd Feature')
    gfssf_pruner = GFSSFPruner(
        X_df, y_df, X_df, y, [feature_weak], feature_strong)

    redunant_features = gfssf_pruner.prune()
    assert feature_weak in redunant_features, \
        'Noisy features should be pruned'


def test_gfssf_pruner_keep_relevant(sample_data):
    X_df, y_df, y = sample_data

    feature_1 = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    feature_2 = Feature(
        input='Z_0',
        transformer=IdentityTransformer(),
        source='2nd Feature')
    gfssf_pruner = GFSSFPruner(
        X_df, y_df, X_df, y, [feature_1], feature_2)

    redunant_features = gfssf_pruner.prune()
    assert feature_1 not in redunant_features, \
        'Still relevant features should be pruned'


@pytest.mark.xfail
def test_gfssf_pruner_prune_irrelevant_features(sample_data):
    X_df, y_df, y = sample_data

    feature_1 = Feature(
        input='Z_0',
        transformer=IdentityTransformer(),
        source='1st Feature')
    feature_2 = Feature(
        input='A_0',
        transformer=IdentityTransformer(),
        source='2nd Feature')
    gfssf_pruner = GFSSFPruner(
        X_df, y_df, X_df, y, [feature_1], feature_2)

    redunant_features = gfssf_pruner.prune()
    assert feature_1 in redunant_features, \
        'Irrelevant features should be pruned'



import ast
import inspect
import platform
from types import FunctionType
from typing import Iterator, Set

import black
from funcy import notnone


def is_valid_python(code: str) -> bool:
    """Check if the string is valid python code

    Args:
        code: code as string
    """
    try:
        ast.parse(code)
    except SyntaxError:
        return False
    else:
        return True


def get_target_python_versions() -> Set[black.TargetVersion]:
    major, minor, _ = platform.python_version_tuple()
    pyversion = f'py{major}{minor}'
    return {black.TargetVersion[pyversion.upper()]}


def blacken_code(code: str) -> str:
    """Format code content using Black

    Args:
        code: code as string
    """
    target_versions = get_target_python_versions()
    line_length = black.DEFAULT_LINE_LENGTH
    string_normalization = True

    mode = black.FileMode(
        target_versions=target_versions,
        line_length=line_length,
        string_normalization=string_normalization,
    )

    try:
        return black.format_file_contents(code, fast=False, mode=mode)
    except black.NothingChanged:
        return code


def get_source(f: FunctionType) -> str:
    """Extract the source code from a given function.

    Recursively extracts the source code for all local functions called by
    given function. The resulting source code is encoded in utf-8.

    Known limitation: Cannot use on function defined interactively in normal
    Python terminal. (Functions defined interactively in IPython are okay.)

    Args:
        f: function

    Returns:
        extracted source code

    Raises:
        NotImplementedError: if the function was defined interactively
    """
    lines = _get_source(f, f.__code__.co_filename, f.__name__, set())
    code = '\n'.join(filter(notnone, lines))

    # post-processing
    if isinstance(code, bytes):
        code = code.decode("utf-8")
    code = blacken_code(code)
    return code


def _get_source(
    f: FunctionType, filename: str, symbolname: str, seen: set
) -> Iterator[str]:
    if f in seen:
        return []

    seen.add(f)

    # known limitation: cannot use from stdin
    if f.__code__.co_filename == '<stdin>':
        raise NotImplementedError(
            'Cannot use \'get_source\' on function defined interactively.')

    # if f was not defined in the same file, return code to import it
    # TODO
    if f.__code__.co_filename != filename:
        return []

    # get source of referenced symbols
    for symbolname in f.__code__.co_names:
        obj = f.__globals__.get(symbolname)
        if obj and inspect.isfunction(obj):
            yield from _get_source(obj, filename, symbolname, seen)

    # get source of self
    yield inspect.getsource(f)



from typing import Callable, Collection, NamedTuple, Tuple, cast

import numpy as np
import pandas as pd
from funcy import iterable
from sklearn_pandas import DataFrameMapper
from stacklog import stacklog

# n.b. cannot import Feature here bc of circular import
import ballet.feature
from ballet.eng import BaseTransformer
from ballet.eng.misc import NullTransformer
from ballet.util.log import logger
from ballet.util.typing import OneOrMore


class FeatureEngineeringPipeline(DataFrameMapper):
    """Feature engineering pipeline

    Args:
        features: feature or list of features
    """

    def __init__(self, features: OneOrMore['ballet.feature.Feature']):
        if not features:
            _features = [
                ballet.feature.Feature(input=[],
                                       transformer=NullTransformer())
            ]
        elif not iterable(features):
            features = cast(ballet.feature.Feature, features)
            _features = [features, ]
        else:
            features = cast(Collection[ballet.feature.Feature], features)
            _features = list(features)

        self._ballet_features = _features

        super().__init__(
            [t.as_input_transformer_tuple() for t in _features],
            input_df=True)

    @property
    def ballet_features(self) -> Collection['ballet.feature.Feature']:
        return self._ballet_features


class EngineerFeaturesResult(NamedTuple):
    X_df: pd.DataFrame
    features: Collection['ballet.feature.Feature']
    pipeline: FeatureEngineeringPipeline
    X: np.ndarray
    y_df: pd.DataFrame
    encoder: BaseTransformer
    y: np.ndarray


def make_engineer_features(
    pipeline: FeatureEngineeringPipeline,
    encoder: BaseTransformer,
    load_data: Callable[..., Tuple[pd.DataFrame, pd.DataFrame]],
) -> Callable[[pd.DataFrame, pd.DataFrame], EngineerFeaturesResult]:
    features = pipeline.ballet_features

    @stacklog(logger.info, 'Building features and target')
    def engineer_features(
        X_df: pd.DataFrame = None, y_df: pd.DataFrame = None
    ) -> EngineerFeaturesResult:
        """Build features and target

        Args:
            X_df: raw variables
            y_df: raw target

        Returns:
            build result
        """
        if X_df is None or y_df is None:
            _X_df, _y_df = load_data()
        if X_df is None:
            X_df = _X_df
        if y_df is None:
            y_df = _y_df

        pipeline = FeatureEngineeringPipeline(features)
        X = pipeline.fit_transform(X_df, y=y_df)
        y = encoder.fit_transform(y_df)

        return EngineerFeaturesResult(
            X_df=X_df, features=features, pipeline=pipeline, X=X,
            y_df=y_df, encoder=encoder, y=y)

    return engineer_features



import logging
from typing import Optional, Union

import ballet

TRACE = 7
SIMPLE_LOG_FORMAT = r'%(levelname)s - %(message)s'
DETAIL_LOG_FORMAT = r'[%(asctime)s] {%(name)s: %(filename)s:%(lineno)d} %(levelname)s - %(message)s'  # noqa E501

logging.addLevelName(TRACE, 'TRACE')
logger = logging.getLogger(ballet.__name__)
_handler = None


def enable(logger: Union[str, logging.Logger] = logger,
           level: Union[str, int] = logging.INFO,
           format: str = SIMPLE_LOG_FORMAT,
           echo: bool = True):
    """Enable simple console logging for this module

    Args:
        logger: logger to enable. Defaults to ballet logger.
        level: logging level, either as string (``'INFO'``) or as int
            (``logging.INFO`` or ``20``). Defaults to ``'INFO'``.
        format : logging format. Defaults to :py:const:`SIMPLE_LOG_FORMAT`.
        echo: Whether to log a message at the configured log level to
            confirm that logging is enable. Defaults to True.
    """
    if isinstance(logger, str):
        logger = logging.getLogger(logger)

    global _handler
    if _handler is None:
        _handler = logging.StreamHandler()
        _handler.setFormatter(logging.Formatter(format))

    if isinstance(level, str):
        level = level.upper()
    levelInt: int = logging._checkLevel(level)  # type: ignore
    levelName: str = logging.getLevelName(levelInt)

    logger.setLevel(levelName)
    # _handler.setLevel(levelName)  # might defeat the point

    if _handler not in logger.handlers:
        logger.addHandler(_handler)

    if echo:
        logger.log(levelInt, f'Logging enabled at level {levelName}.')


class LevelFilter:
    """Logging filter for log records at an exact level

    Emits only log records that have the exact level that is given to the
    filter.

    Args:
        level: numeric logging level to filter

    Example usage:

        >>> debugFilter = LevelFilter(logging.DEBUG)
        >>> logger.addFilter(debugFilter)
    """

    def __init__(self, level: int):
        self.__level = level

    def filter(self, logRecord: logging.LogRecord) -> int:
        # Is the specified record to be logged? Returns zero for no, nonzero
        # for yes.
        return 1 if logRecord.levelno == self.__level else 0


class LoggingContext:
    """Logging context manager

    Configure the given logger to emit messages at and above the configured
    logging level and using the configured logging handler. Useful to
    temporarily set a lower (or higher) log level or to temporarily add a
    local handler. After the context exits, the original state of the logger
    will be restored.

    If a handler is not provided, messages will continue to be logged to the
    existing handlers attached to the logger. If the logger's level is
    lowered, the handlers may

    Args:
        logger: logger
        level: string or numeric logging level
        handler: log handler if not is already configured
        close: whether to close the handler after context exits.
            Defaults to True.

    Example usage:

        >>> with LoggingContext(logger, level='DEBUG'):
        ...     logger.debug('some message')

    Source: <https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging>
    """  # noqa E501

    def __init__(self,
                 logger: logging.Logger,
                 level: Union[str, int, None] = None,
                 handler: Optional[logging.Handler] = None,
                 close: bool = True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler is not None:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler is not None:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        return None



import numpy as np

from ballet.compat import SimpleImputer
from ballet.eng.base import BaseTransformer
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.util import has_nans
from ballet.validation.common import check_from_class
from ballet.validation.feature_api.checks import (
    CanDeepcopyCheck, CanTransformCheck, FeatureApiCheck,
    HasCorrectInputTypeCheck, HasCorrectOutputDimensionsCheck,
    NoMissingValuesCheck,)

from ..util import FragileTransformer


def test_good_feature(sample_data):
    feature = Feature(
        input='size',
        transformer=SimpleImputer(),
    )

    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert valid
    assert len(failures) == 0


def test_bad_feature_input(sample_data):
    # bad input
    feature = Feature(
        input=3,
        transformer=SimpleImputer(),
    )
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert not valid
    assert HasCorrectInputTypeCheck.__name__ in failures


def test_bad_feature_transform_errors(sample_data):
    # transformer throws errors
    feature = Feature(
        input='size',
        transformer=FragileTransformer(
            (lambda x: True, ), (RuntimeError, ))
    )
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert not valid
    assert CanTransformCheck.__name__ in failures


def test_bad_feature_wrong_transform_length(sample_data):
    class _WrongLengthTransformer(BaseTransformer):
        def transform(self, X, **transform_kwargs):
            new_shape = list(X.shape)
            new_shape[0] += 1
            output = np.arange(np.prod(new_shape)).reshape(new_shape)
            return output

    # doesn't return correct length
    feature = Feature(
        input='size',
        transformer=_WrongLengthTransformer(),
    )
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert not valid
    assert HasCorrectOutputDimensionsCheck.__name__ in failures


def test_bad_feature_deepcopy_fails(sample_data):
    class _CopyFailsTransformer(IdentityTransformer):
        def __deepcopy__(self, memo):
            raise RuntimeError
    feature = Feature(
        input='size',
        transformer=_CopyFailsTransformer(),
    )
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert not valid
    assert CanDeepcopyCheck.__name__ in failures


def test_producing_missing_values_fails(sample_data):
    assert has_nans(sample_data.X)
    feature = Feature(
        input='size',
        transformer=IdentityTransformer()
    )
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, sample_data.X, sample_data.y)
    assert not valid
    assert NoMissingValuesCheck.__name__ in failures



from unittest.mock import patch

import funcy
import pytest
from git import GitCommandError

import ballet.exc
import ballet.templating
import ballet.update
from ballet.project import DEFAULT_CONFIG_NAME
from ballet.update import DEFAULT_BRANCH, TEMPLATE_BRANCH
from ballet.util import work_in

# Utility methods -------------------------------------------------------------


def _run_ballet_update_template(d, project_slug, **kwargs):
    with work_in(d.joinpath(project_slug)):
        ballet.update.update_project_template(**kwargs)


def check_remotes(repo, expected_remotes=None):
    def get_names(remotes):
        return funcy.lpluck_attr('name', remotes)

    if expected_remotes is None:
        expected_remote_names = ['origin']
    else:
        expected_remote_names = get_names(expected_remotes)

    actual_remotes = repo.remotes
    actual_remote_names = get_names(actual_remotes)

    assert actual_remote_names == expected_remote_names


def check_branch(repo, expected_branch=DEFAULT_BRANCH):
    actual_branch = repo.head.ref.name
    assert actual_branch == expected_branch


def check_commit_message_on_template(repo):
    expected_commit_message = \
        ballet.update._make_template_branch_merge_commit_message()
    commit = repo.heads[TEMPLATE_BRANCH].commit
    check_commit_message(commit, expected_commit_message)


def check_commit_message_on_master(repo):
    expected_commit_message = f'Merge branch {TEMPLATE_BRANCH!r}'
    commit = repo.head.commit
    check_commit_message(commit, expected_commit_message)


def check_commit_message(commit, expected_commit_message):
    actual_commit_message = commit.message.strip()
    assert actual_commit_message == expected_commit_message


def check_modified_file(path, content):
    with path.open('r') as f:
        lines = f.readlines()
        last_line = lines[-1]
        assert content in last_line


# Tests begin------------------------------------------------------------------

@pytest.mark.slow
def test_update_after_change_in_template(quickstart, project_template_copy):
    """Test update after changing a file in the template

    We expect the project-template branch to be updated and then successfully
    merged into master branch without merge conflicts, and for the changes from
    the template to be reflected in the file on master.

    Setup:

        # quickstart
        # copy project template for modification

    Replicate the following script:

        $ echo '\nfoo: bar' >> /path/to/project_template/ballet.yml
        $ ballet-update-template

    Then test the following:
        - on branch master
        - commit msg on master matches expected msg
        - foo: bar is the last line of ballet.yml
    """
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug
    repo = quickstart.repo

    template_dir = project_template_copy

    modified_file_path = tempdir.joinpath(project_slug, DEFAULT_CONFIG_NAME)

    # need to indent
    new_content = 'foo: bar'

    # add foo: bar to project template
    p = template_dir.joinpath('{{cookiecutter.project_slug}}',
                              DEFAULT_CONFIG_NAME)
    with p.open('a') as f:
        f.write('\n')
        f.write(' ' * 2)  # indent under `default` key
        f.write(new_content)
        f.write('\n')

    # run ballet-update-template
    _run_ballet_update_template(tempdir, project_slug)

    # assert on branch master
    check_branch(repo)

    # assert commit message is automatically generated
    check_commit_message_on_master(repo)
    check_commit_message_on_template(repo)

    # assert foo: bar is in ballet.yml
    check_modified_file(modified_file_path, new_content)

    # assert no new remotes
    check_remotes(repo)


@pytest.mark.slow
def test_update_after_change_in_project(quickstart):
    """Test update after changing a file in the project itself

    We expect the updated template to be seamlessly merged, with no merge
    conflicts, and the change to continue to be reflected in the file.

    Setup:

        # quickstart

    Replicate the following script

        $ echo '\nfoo: bar' >> ./ballet.yml
        $ git add ./ballet.yml
        $ git commit -m 'Add "foo: bar" to ballet.yml'
        $ ballet-update-template

    Then test the following:
        - on branch master
        - no new commits to project-template branch
        - no new commits to master branch
        - foo: bar is the last line of ballet.yml
    """
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug
    repo = quickstart.repo

    expected_template_commit = repo.branches[TEMPLATE_BRANCH].commit

    modified_file_path = tempdir.joinpath(project_slug, DEFAULT_CONFIG_NAME)
    new_content = 'foo: bar'
    # add foo: bar
    with modified_file_path.open('a') as f:
        f.write('\n')
        f.write(new_content)
        f.write('\n')

    # commit
    repo.git.add(DEFAULT_CONFIG_NAME)
    repo.git.commit(
        m=f'Add \'{new_content}\' to {DEFAULT_CONFIG_NAME}')

    expected_master_commit = repo.head.commit

    # run ballet-update-template
    _run_ballet_update_template(tempdir, project_slug)

    # assert on branch master
    check_branch(repo)

    # assert no new remotes
    check_remotes(repo)

    # assert no new commit/changes
    actual_master_commit = repo.head.commit
    assert actual_master_commit == expected_master_commit

    actual_template_commit = repo.branches[TEMPLATE_BRANCH].commit
    assert actual_template_commit == expected_template_commit

    # assert foo: bar is in ballet.yml
    check_modified_file(modified_file_path, new_content)


@pytest.mark.slow
def test_update_after_conflicting_changes(quickstart, project_template_copy):
    """Test what happens when there are conflicting changes

    We expect the merge between project-template and master branches to fail in
    some way, whether that is a Python error or just a message to the user, it
    is not decided yet.

    """
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug
    repo = quickstart.repo

    # add foo: bar
    with tempdir.joinpath(project_slug, DEFAULT_CONFIG_NAME).open('a') as f:
        f.write('\nfoo: bar\n')

    # commit
    repo.git.add(DEFAULT_CONFIG_NAME)
    repo.git.commit(m='Add "foo: bar" to {}')

    # add foo: qux to project template
    template_dir = project_template_copy
    p = template_dir.joinpath('{{cookiecutter.project_slug}}',
                              DEFAULT_CONFIG_NAME)
    with p.open('a') as f:
        f.write('\nfoo: qux\n')

    # run ballet-update-template -- this should raise an error because there
    # should be a merge conflict
    with pytest.raises(GitCommandError):
        _run_ballet_update_template(tempdir, project_slug)

    # assert on branch master
    check_branch(repo)

    # assert no new remotes?
    check_remotes(repo)


@pytest.mark.slow
def test_update_after_no_changes(quickstart):
    """Test that if there are no changes to project template, nothing happens.

    Currently, this test fails because an empty merge commit is created and
    an empty project template update commit is created.

    Setup:

        # quickstart

    Replicate the following script

        $ ballet-update-template

    Then test the following:
        - on branch master
        - no new remotes
        - no new commit/changes
    """
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug
    repo = quickstart.repo

    expected_master_commit = repo.head.commit
    expected_template_commit = repo.branches[TEMPLATE_BRANCH].commit

    _run_ballet_update_template(tempdir, project_slug)

    # assert on branch master
    check_branch(repo)

    # assert no new remotes
    check_remotes(repo)

    # assert no new commit/changes
    actual_master_commit = repo.head.commit
    assert actual_master_commit == expected_master_commit

    actual_template_commit = repo.branches[TEMPLATE_BRANCH].commit
    assert actual_template_commit == expected_template_commit


@pytest.mark.slow
def test_update_fails_with_dirty_repo(quickstart):
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug
    with tempdir.joinpath(project_slug, DEFAULT_CONFIG_NAME).open('a') as f:
        f.write('\nfoo: bar\n')

    with pytest.raises(ballet.exc.BalletError, match='uncommitted changes'):
        _run_ballet_update_template(tempdir, project_slug)


@pytest.mark.slow
def test_update_fails_with_missing_project_template_branch(quickstart):
    tempdir = quickstart.tempdir
    repo = quickstart.repo
    project_slug = quickstart.project_slug

    repo.delete_head(TEMPLATE_BRANCH)

    with pytest.raises(ballet.exc.ConfigurationError):
        _run_ballet_update_template(tempdir, project_slug)


@pytest.mark.slow
@patch('git.Remote.push')
def test_update_push(mock_push, quickstart, project_template_copy):
    # TODO(mjs)
    # make this unit tests instead
    # update this test to test more behaviors of the push
    # test failure by using non-existent remote locally
    # test success by mocking bare repo locally
    tempdir = quickstart.tempdir
    project_slug = quickstart.project_slug

    # update the project template so the update command runs to completion
    new_content = 'foo: bar'
    template_dir = project_template_copy
    p = template_dir.joinpath(
        '{{cookiecutter.project_slug}}', DEFAULT_CONFIG_NAME)
    with p.open('a') as f:
        f.write('\n')
        f.write(new_content)
        f.write('\n')

    _run_ballet_update_template(tempdir, project_slug, push=True)

    mock_push.assert_called_once()



import pathlib
import random
import shlex
import subprocess
import sys
from unittest.mock import ANY, PropertyMock, patch

import git
import pytest

from ballet.project import Project, detect_github_username, load_config
from ballet.util.testing import seeded


@patch('ballet.project.load_config_in_dir')
def test_load_config(mock_load_config_in_dir):
    path = pathlib.Path(__file__)
    load_config(path=path)
    mock_load_config_in_dir.assert_called_once_with(path)


@patch('ballet.project.load_config_in_dir')
def test_load_config_detect(mock_load_config_in_dir):
    path = pathlib.Path(__file__)
    load_config()
    mock_load_config_in_dir.assert_called_once_with(path)


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason='capture_output added in py37, not worth it to add compat')
@patch('ballet.project.load_config_in_dir')
def test_load_config_dash_c(mock_load_config_in_dir, quickstart):
    # note that this tests python -c 'cmd' which is different from the repl!
    pycmd = 'from ballet.project import load_config; config=load_config()'
    cmd = f"{sys.executable} -c '{pycmd}'"
    result = subprocess.run(
        shlex.split(cmd), cwd=quickstart.project.path, capture_output=True)
    assert not result.returncode, result.stderr


@pytest.mark.xfail
def test_load_config_at_path():
    raise NotImplementedError


@pytest.mark.xfail
def test_load_config_in_dir():
    raise NotImplementedError


@patch('ballet.project.Project.repo', new_callable=PropertyMock)
def test_detect_github_username_config(mock_project_repo):
    expected_username = 'Foo Bar'

    # output of project.repo.config_reader().get_value(...)
    mock_get_value = (mock_project_repo
                      .return_value
                      .config_reader
                      .return_value
                      .get_value)
    mock_get_value.return_value = expected_username

    project = Project(None)
    username = detect_github_username(project)
    assert expected_username == username

    mock_get_value.assert_called_with('github', 'user', default=ANY)


@pytest.fixture
def commit_object():
    with seeded(17):
        # 20-byte sha1
        k = 20
        bits = random.getrandbits(k * 8)
        data = bits.to_bytes(k, sys.byteorder)  # in py39, can use randbytes(k)

        repo = None
        commit = git.Commit(repo, data)
        yield commit


@patch('ballet.project.Project.repo', new_callable=PropertyMock)
def test_project_version(mock_repo, commit_object):
    mock_repo.return_value.head.commit = commit_object
    project = Project(None)
    version = project.version
    assert isinstance(version, str)



from ballet.eng.base import *
from ballet.eng.misc import *
from ballet.eng.missing import *
from ballet.eng.ts import *

# needed for sphinx
from ballet.eng.base import __all__ as _base_all
from ballet.eng.misc import __all__ as _misc_all
from ballet.eng.missing import __all__ as _missing_all
from ballet.eng.ts import __all__ as _ts_all
__all__ = (*_base_all, *_misc_all, *_missing_all, *_ts_all)



import logging
import random

import pytest

from ballet.util.log import LevelFilter, LoggingContext, enable


@pytest.fixture
def logger():
    name = str(random.randint(0, 1 << 10))
    return logging.getLogger(name)


@pytest.mark.parametrize('level', [logging.INFO, 'CRITICAL'])
def test_enable(logger, caplog, level):
    with caplog.at_level(level, logger=logger.name):
        enable(logger, level, echo=True)
    assert 'enabled' in caplog.text


def test_level_filter_matches(logger, caplog):
    enable(logger, level='DEBUG', echo=False)
    logger.addFilter(LevelFilter(logging.CRITICAL))

    # does log message at level CRITICAL
    with caplog.at_level(logging.CRITICAL, logger=logger.name):
        logger.critical('msg')

    assert caplog.text


def test_level_filter_not_matches(logger, caplog):
    enable(logger, level='DEBUG', echo=False)
    logger.addFilter(
        LevelFilter(logging.DEBUG))

    # does *not* log message at level INFO > DEBUG
    with caplog.at_level(logging.INFO, logger=logger.name):
        logger.info('msg')

    assert not caplog.text


def test_logging_context(logger, caplog):
    enable(logger, level='DEBUG', echo=False)
    with caplog.at_level(logging.DEBUG, logger=logger.name):
        with LoggingContext(logger, level='INFO'):
            logger.debug('msg')
    assert not caplog.text



from typing import Callable, Collection, TypeVar, Union

import ballet.eng  # noqa
from ballet.compat import PathLike

T = TypeVar('T')
OneOrMore = Union[T, Collection[T]]
Pathy = Union[str, PathLike]
TransformerLike = Union[Callable, 'ballet.eng.BaseTransformer', None]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md', encoding='utf-8') as history_file:
    history = history_file.read()

requirements = [
    'black',
    'cookiecutter',
    'Click >= 6.0',
    'dataclasses; python_version == "3.6"',
    'dill',
    'dynaconf ~= 3.0',
    'funcy >= 1.14',
    'gitpython',
    'h5py',
    'numpy',
    'packaging',
    'pandas ~= 1.0',
    'pygithub ~= 1.54',
    'pyyaml',
    'requests',
    'scikit_learn >=0.20, <0.23; python_version == "3.6"',
    'scikit_learn >= 0.20; python_version > "3.6"',
    'scipy',
    'sklearn_pandas ~= 1.0',
    'stacklog',
]

extras = {
    'category_encoders': ['category_encoders >= 2.2.2'],
    'feature_engine': ['feature_engine ~= 1.0'],
    'featuretools': ['featuretools_sklearn_transformer >= 0.1'],
    'skits': ['skits >= 0.1.2'],
    'tsfresh': ['tsfresh >= 0.16'],
}
extras['all'] = [dep for deps in extras.values() for dep in deps]

test_requirements = [
    'coverage >= 4.5.1',
    'pytest >= 6',
    'pytest-cov >= 2.6',
    'pytest-virtualenv >= 1.7.0',
    'tox >= 2.9.1',
    'responses >= 0.13.2',
]

development_requirements = [
    # general
    'bump2version >= 1',
    'pip >= 9.0.1',
    'watchdog[watchmedo] >= 0.8.3',
    'invoke >= 1.4',
    'mypy >= 0.812',

    # docs
    'm2r2 >= 0.2.5',
    'sphinx ~= 3.0',
    'sphinx_rtd_theme >= 0.2.4',
    'sphinx-click >= 2.2',
    'sphinx-autodoc-typehints >= 1.11',
    'rstcheck',

    # style check
    'flake8 >= 3.5.0',
    'isort >= 5.0',

    # fix style issues
    'autopep8 >= 1.3.5',

    # distribute on PyPI
    'twine >= 1.10.0',
    'wheel >= 0.30.0',
]

setup(
    author="Micah Smith",
    author_email='micahs@mit.edu',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description='Core functionality for lightweight, collaborative data science projects',
    entry_points={
        'console_scripts': [
            'ballet=ballet.cli:cli',
        ],
    },
    extras_require={
        'test': test_requirements,
        'dev': development_requirements + test_requirements,
        **extras,
    },
    install_requires=requirements,
    license='MIT license',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='ballet',
    name='ballet',
    packages=find_packages(include=['ballet', 'ballet.*']),
    python_requires='>=3.6.1',
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ballet/ballet',
    version='0.17.0',
    zip_safe=False,
)



from typing import Union

import pandas as pd

from ballet import logger
from ballet.feature import Feature
from ballet.validation.common import (
    check_from_class, subsample_data_for_validation,)
from ballet.validation.feature_api.checks import FeatureApiCheck


def validate_feature_api(
    feature: Feature,
    X_df: pd.DataFrame,
    y_df: Union[pd.DataFrame, pd.Series],
    subsample: bool,
    log_advice: bool = False,
) -> bool:
    logger.debug(f'Validating feature {feature!r}')
    if subsample:
        X_df, y_df = subsample_data_for_validation(X_df, y_df)
    valid, failures, advice = check_from_class(
        FeatureApiCheck, feature, X_df, y_df)
    if valid:
        logger.info('Feature is valid')
    else:
        if log_advice:
            logger.info(
                'Feature is NOT valid; here is some advice for resolving the '
                'feature API issues.')
            for failure, advice_item in zip(failures, advice):
                logger.info(f'{failure}: {advice_item}')
        else:
            logger.info(f'Feature is NOT valid; failures were {failures}')

    return valid



import json
import pathlib
import tempfile
from textwrap import dedent
from typing import Optional
from unittest.mock import patch

import funcy
import git
import packaging.version
import requests
from cookiecutter.prompt import prompt_for_config
from git import GitCommandError

import ballet
from ballet.exc import BalletError, ConfigurationError
from ballet.project import Project
from ballet.templating import render_project_template
from ballet.util.git import DEFAULT_BRANCH, push_branches_to_remote
from ballet.util.log import logger
from ballet.util.typing import Pathy

PROJECT_CONTEXT_PATH = (
    pathlib.Path(__file__).resolve().parent.joinpath(
        'templates',
        'project_template',
        'cookiecutter.json'))
CONTEXT_FILE_NAME = '.cookiecutter_context.json'
TEMPLATE_BRANCH = 'project-template'
PYPI_PROJECT_JSON_URL = 'https://pypi.org/pypi/{project}/json'


def _get_latest_project_version_string(project: str) -> Optional[str]:
    """Get the latest version of a project according to the PyPI Warehouse API

    For context, the `sampleproject` project returns 10KB of data.

    Returns:
        latest version of `project` or None if something went wrong
    """
    url = PYPI_PROJECT_JSON_URL.format(project=project)
    response = requests.get(url)
    releases = response.json()['releases'].keys()
    return max(releases, key=packaging.version.parse)


@funcy.ignore(Exception)
def _get_latest_ballet_version_string() -> Optional[str]:
    """Get the latest version of ballet according to the PyPI Warehouse API

    This returns on the order of 50KB of data.

    Returns:
        latest version of ballet or None if something went wrong
    """
    return _get_latest_project_version_string(ballet.__name__)


def _check_for_updated_ballet() -> Optional[str]:
    """Return the version of an updated ballet if it is available

    Returns:
        the latest version of ballet available or None if the latest version
        is the same as the currently-installed version
    """
    latest = _get_latest_ballet_version_string()
    current = ballet.__version__
    parse = packaging.version.parse
    if latest and parse(latest) > parse(current):
        return latest
    else:
        return None


def _warn_of_updated_ballet(latest: Optional[str]):
    if latest is not None:
        current = ballet.__version__
        msg = dedent(
            f'''
            A new version of ballet is available: v{latest}
            - you currently have ballet v{current}
            - if you don't update ballet, you won't receive project template updates
            - update ballet and then try again:

                $ python -m pip install --upgrade ballet
            '''  # noqa E501
        ).strip()
        logger.warning(msg)


def _make_template_branch_merge_commit_message() -> str:
    version = ballet.__version__
    return f'Merge project template updates from ballet v{version}'


@funcy.silent
def _safe_delete_remote(repo: git.Repo, name: str):
    repo.delete_remote(name)


def _render_project_template(
    cwd: pathlib.Path,
    tempdir: Pathy,
    project_template_path: Optional[Pathy] = None
) -> str:
    tempdir = pathlib.Path(tempdir)
    context = _get_full_context(cwd)

    # don't dump replay files to home directory.
    with patch('cookiecutter.main.dump'):
        return render_project_template(
            project_template_path=project_template_path,
            no_input=True,
            extra_context=context,
            output_dir=tempdir)


def _get_full_context(cwd: pathlib.Path) -> dict:
    # load the context stored within the project repository
    context_path = cwd.joinpath(CONTEXT_FILE_NAME)
    if context_path.exists():
        with context_path.open('r') as f:
            context = json.load(f)
    else:
        raise FileNotFoundError(
            f'Could not find \'{CONTEXT_FILE_NAME}\', are you in a ballet '
            'project repo?')

    # find out if there are any new keys to prompt for
    with PROJECT_CONTEXT_PATH.open('r') as f:
        new_context = json.load(f)
    new_keys = set(new_context) - set(context['cookiecutter'])
    if new_keys:
        new_context_config = {'cookiecutter': funcy.project(new_context,
                                                            new_keys)}
        new_context = prompt_for_config(new_context_config)
        context['cookiecutter'].update(new_context)

    return context['cookiecutter']


def _log_recommended_reinstall():
    logger.info(
        'After a successful project template update, try re-installing the\n'
        'project in case the project template requires any different \n'
        'dependencies than what you have installed:\n'
        '\n'
        '    $ invoke install')


def update_project_template(push: bool = False,
                            project_template_path: Optional[Pathy] = None):
    """Update project with updates to upstream project template

    The update is fairly complicated and proceeds as follows:

    1. Load project: user must run command from master branch and ballet
       must be able to detect the project-template branch
    2. Load the saved cookiecutter context from disk
    3. Render the project template into a temporary directory using the
       saved context, *prompting the user if new keys are required*. Note
       that the project template is simply loaded from the data files of the
       installed version of ballet. Note further that by the project
       template's post_gen_hook, a new git repo is initialized [in the
       temporary directory] and files are committed.
    4. Add the temporary directory as a remote and merge it into the
       project-template branch, favoring changes made to the upstream template.
       Any failure to merge results in an unrecoverable error.
    5. Merge the project-template branch into the master branch. The user is
       responsible for merging conflicts and they are given instructions to
       do so and recover.
    6. If applicable, push to master.

    Args:
        push: whether to push updates to remote, defaults to False
        project_template_path: an override for the path to the
            project template
    """
    cwd = pathlib.Path.cwd().resolve()

    # get ballet project info -- must be at project root directory with a
    # ballet.yml file.
    try:
        project = Project.from_path(cwd)
    except ConfigurationError:
        raise ConfigurationError('Must run command from project root.')

    repo = project.repo
    original_head = repo.head.commit.hexsha[:7]

    if repo.is_dirty():
        raise BalletError(
            'Can\'t update project template with uncommitted changes. '
            'Please commit your changes and try again.')

    if repo.head.ref.name != DEFAULT_BRANCH:
        raise ConfigurationError(
            f'Must run command from branch {DEFAULT_BRANCH}')

    if TEMPLATE_BRANCH not in repo.branches:
        raise ConfigurationError(
            f'Could not find \'{TEMPLATE_BRANCH}\' branch.')

    # check for upstream updates to ballet
    new_version = _check_for_updated_ballet()
    if new_version:
        _warn_of_updated_ballet(new_version)

    with tempfile.TemporaryDirectory() as _tempdir:
        tempdir = pathlib.Path(_tempdir)

        # cookiecutter returns path to the resulting project dir
        logger.debug(f'Re-rendering project template at {tempdir}')
        updated_template = _render_project_template(
            cwd, tempdir, project_template_path=project_template_path)
        updated_repo = git.Repo(updated_template)

        # tempdir is a randomly-named dir suitable for a random remote name
        # to avoid conflicts
        remote_name = tempdir.name

        remote = repo.create_remote(
            remote_name, updated_repo.working_tree_dir)
        remote.fetch()

        repo.heads[TEMPLATE_BRANCH].checkout()
        try:
            logger.debug('Merging re-rendered template to project-template '
                         'branch')
            repo.git.merge(
                remote_name + '/' + DEFAULT_BRANCH,
                allow_unrelated_histories=True,
                strategy_option='theirs',
                squash=True,
            )
            if not repo.is_dirty():
                logger.info('No updates to template -- done.')
                return
            commit_message = _make_template_branch_merge_commit_message()
            logger.debug(f'Committing updates: {commit_message}')
            repo.git.commit(m=commit_message)
        except GitCommandError:
            logger.critical(
                f'Could not merge changes into {TEMPLATE_BRANCH} branch, '
                f'update failed')
            raise
        finally:
            _safe_delete_remote(repo, remote_name)
            logger.debug('Checking out master branch')
            repo.heads[DEFAULT_BRANCH].checkout()

    try:
        logger.debug('Merging project-template branch into master')
        repo.git.merge(TEMPLATE_BRANCH, no_ff=True)
    except GitCommandError as e:
        if 'merge conflict' in str(e).lower():
            logger.critical(dedent(
                f'''
                Update failed due to a merge conflict.
                Fix conflicts, and then complete merge manually:
                    $ git add .
                    $ git commit --no-edit
                Otherwise, abandon the update:
                    $ git reset --merge {original_head}
                '''
            ).strip())
        raise

    if push:
        repo = project.repo
        remote_name = project.config.get('github.remote')
        branches = [DEFAULT_BRANCH, TEMPLATE_BRANCH]
        push_branches_to_remote(repo, remote_name, branches)

    _log_recommended_reinstall()



import numpy as np
import pandas as pd
import pytest
from scipy.special import boxcox1p

import ballet.eng.misc
from ballet.util.testing import assert_array_almost_equal, assert_array_equal


def test_value_replacer():
    trans = ballet.eng.misc.ValueReplacer(0.0, -99)
    data = pd.DataFrame([0, 0, 0, 0, 1, 3, 7, 11, -7])
    expected_result = pd.DataFrame([-99, -99, -99, -99, 1, 3, 7, 11, -7])

    result = trans.fit_transform(data)
    pd.util.testing.assert_frame_equal(result, expected_result)


def test_box_cox_transformer():
    threshold = 0.0
    lmbda = 0.0
    trans = ballet.eng.misc.BoxCoxTransformer(
        threshold=threshold, lmbda=lmbda)

    skewed = [0., 0., 0., 0., 1.]
    unskewed = [0., 0., 0., 0., 0.]

    exp_skew_res = boxcox1p(skewed, lmbda)
    exp_unskew_res = unskewed

    # test on DF, one skewed column
    df = pd.DataFrame()
    df['skewed'] = skewed
    df['unskewed'] = unskewed
    df_res = trans.fit_transform(df)
    assert isinstance(df_res, pd.DataFrame)
    assert 'skewed' in df_res.columns
    assert 'unskewed' in df_res.columns
    assert_array_almost_equal(df_res['skewed'], exp_skew_res)

    # test on DF, no skewed columns
    df_unskewed = pd.DataFrame()
    df_unskewed['col1'] = unskewed
    df_unskewed['col2'] = unskewed
    df_unskewed_res = trans.fit_transform(df_unskewed)
    assert isinstance(df_unskewed_res, pd.DataFrame)
    assert 'col1' in df_unskewed_res.columns
    assert 'col2' in df_unskewed_res.columns
    assert_array_equal(df_unskewed_res['col1'], exp_unskew_res)
    assert_array_equal(df_unskewed_res['col2'], exp_unskew_res)

    # test on skewed Series
    ser_skewed = pd.Series(skewed)
    ser_skewed_res = trans.fit_transform(ser_skewed)
    assert isinstance(ser_skewed_res, pd.Series)
    assert_array_almost_equal(ser_skewed_res, exp_skew_res)

    # test on unskewed Series
    ser_unskewed = pd.Series(unskewed)
    ser_unskewed_res = trans.fit_transform(ser_unskewed)
    assert isinstance(ser_unskewed_res, pd.Series)
    assert_array_equal(ser_unskewed_res, exp_unskew_res)

    # test on array
    arr = np.array([unskewed, skewed]).T
    arr_res = trans.fit_transform(arr)
    arr_exp = np.vstack((exp_unskew_res, exp_skew_res)).T
    assert_array_almost_equal(arr_res, arr_exp)


def test_named_framer():
    name = 'foo'

    # good objects
    index = pd.Index([1, 2, 3])
    ser = pd.Series([1, 2, 3])
    df = pd.DataFrame([1, 2, 3])
    arr = np.array([1, 2, 3])

    for obj in [index, ser, df, arr]:
        trans = ballet.eng.misc.NamedFramer(name)
        result = trans.fit_transform(obj)
        assert isinstance(result, pd.DataFrame)
        assert result.shape[1] == 1
        assert result.columns == [name]

    # bad objects -- wrong type
    int_ = 1
    str_ = 'hello'
    for obj in [int_, str_]:
        trans = ballet.eng.misc.NamedFramer(name)
        with pytest.raises(TypeError):
            result = trans.fit_transform(obj)

    # bad objects -- too wide
    wide_df = pd.DataFrame(np.arange(10).reshape(-1, 2))
    wide_arr = np.arange(10).reshape(-1, 2)
    for obj in [wide_df, wide_arr]:
        trans = ballet.eng.misc.NamedFramer(name)
        with pytest.raises(ValueError):
            result = trans.fit_transform(obj)



import funcy
import numpy as np
import pandas as pd
import pytest
import sklearn.preprocessing
from sklearn_pandas.pipeline import TransformerPipeline

from ballet.compat import SimpleImputer
from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.transformer import (
    DelegatingRobustTransformer, get_transformer_primitives,
    make_robust_transformer,)
from ballet.util import asarray2d

from .util import FragileTransformer, FragileTransformerPipeline


@pytest.fixture
def sample_data():
    X_ser = pd.util.testing.makeFloatSeries()
    X_df = X_ser.to_frame()
    X_arr1d = np.asarray(X_ser)
    X_arr2d = np.asarray(X_df)
    y_ser = X_ser.copy()
    y_df = X_df.copy()
    y_arr1d = np.asarray(y_ser)
    y_arr2d = np.asarray(y_df)
    d = {
        'ser': (X_ser, y_ser),
        'df': (X_df, y_df),
        'arr1d': (X_arr1d, y_arr1d),
        'arr2d': (X_arr2d, y_arr2d),
    }
    return d


@pytest.mark.parametrize(
    'input_types, bad_input_checks, catches',
    [
        (
            ('ser', ),
            (lambda x: isinstance(x, pd.Series), ),
            (ValueError, TypeError),
        ),
        (
            ('ser', 'df', ),
            (
                lambda x: isinstance(x, pd.Series),
                lambda x: isinstance(x, pd.DataFrame)
            ),
            (ValueError, TypeError),
        ),
        (
            ('ser', 'df', 'arr1d'),
            (
                lambda x: isinstance(x, pd.Series),
                lambda x: isinstance(x, pd.DataFrame),
                lambda x: isinstance(x, np.ndarray) and x.ndim == 1,
            ),
            (ValueError, TypeError),
        )
    ],
    ids=[
        'ser',
        'df',
        'arr',
    ]
)
@pytest.mark.parametrize(
    'transformer_maker',
    [
        FragileTransformer,
        funcy.partial(FragileTransformerPipeline, 3)
    ]
)
def test_robust_transformer(
    input_types, bad_input_checks, catches, transformer_maker,
    sample_data,
):
    fragile_transformer = transformer_maker(bad_input_checks, catches)
    robust_transformer = DelegatingRobustTransformer(
        transformer_maker(bad_input_checks, catches))

    for input_type in input_types:
        X, y = sample_data[input_type]
        # fragile transformer raises error
        with pytest.raises(catches):
            fragile_transformer.fit_transform(X, y)
        # robust transformer does not raise error
        X_robust = robust_transformer.fit_transform(X, y)
        assert np.array_equal(
            asarray2d(X),
            asarray2d(X_robust)
        )


def test_robust_transformer_sklearn(sample_data):
    Transformers = (
        SimpleImputer,
        sklearn.preprocessing.StandardScaler,
        sklearn.preprocessing.Binarizer,
        sklearn.preprocessing.PolynomialFeatures,
    )
    # some of these input types are bad for sklearn.
    input_types = ('ser', 'df', 'arr1d')
    for Transformer in Transformers:
        robust_transformer = DelegatingRobustTransformer(Transformer())
        for input_type in input_types:
            X, y = sample_data[input_type]
            robust_transformer.fit_transform(X, y=y)


@pytest.mark.parametrize(
    'robust_maker',
    [
        DelegatingRobustTransformer,
        lambda x: make_robust_transformer([x]),
    ]
)
def test_robust_str_repr(robust_maker):
    robust_transformer = robust_maker(IdentityTransformer())
    for func in [str, repr]:
        s = func(robust_transformer)
        assert len(s) > 0


@pytest.mark.parametrize(
    'transformer,expected',
    [
        (
            IdentityTransformer(),
            ['IdentityTransformer'],
        ),
        (
            [IdentityTransformer(), IdentityTransformer()],
            ['IdentityTransformer', 'IdentityTransformer'],
        )
    ]
)
def test_get_transformer_primitives(transformer, expected):
    robust_transformer = make_robust_transformer(transformer)
    primitives = get_transformer_primitives(robust_transformer)
    assert primitives == expected


def test_robust_transformer_desugar():
    """Should be able to "desugar" multiple things into a valid transformer pipeline"""  # noqa
    transformer = [
        None,
        IdentityTransformer(),
        lambda x: x,
        Feature('A', IdentityTransformer()),
        ('A', IdentityTransformer()),
        ('A', [None, IdentityTransformer()]),
    ]
    robust_transformer = make_robust_transformer(transformer)
    assert isinstance(robust_transformer, TransformerPipeline)



import json
import logging
from unittest.mock import Mock, patch

import git

import ballet
import ballet.util.log
from ballet.update import (
    PYPI_PROJECT_JSON_URL, _check_for_updated_ballet,
    _get_latest_ballet_version_string, _get_latest_project_version_string,
    _log_recommended_reinstall, _make_template_branch_merge_commit_message,
    _safe_delete_remote, _warn_of_updated_ballet,)


def test_get_latest_project_version_string(testdatadir, responses):
    with testdatadir.joinpath('sampleproject.json').open('r') as f:
        data = json.load(f)
    url = PYPI_PROJECT_JSON_URL.format(project='sampleproject')
    responses.add(responses.GET, url, json=data)
    expected = '1.2.0'

    actual = _get_latest_project_version_string('sampleproject')

    assert actual == expected


@patch('ballet.update._get_latest_project_version_string')
def test_get_latest_ballet_version_string(mock_latest):
    expected = mock_latest.return_value
    actual = _get_latest_ballet_version_string()
    assert actual == expected


@patch('ballet.update._get_latest_ballet_version_string')
def test_check_for_updated_ballet(mock_latest):
    # obviously this will represent an update from whatever the current
    # version is
    latest = '99999999999.0.0'
    mock_latest.return_value = latest
    expected = latest
    actual = _check_for_updated_ballet()
    assert actual == expected


@patch('ballet.update._get_latest_ballet_version_string')
def test_check_for_updated_ballet_no_updates(mock_latest):
    mock_latest.return_value = ballet.__version__
    expected = None  # no updates available
    actual = _check_for_updated_ballet()
    assert actual == expected


def test_warn_of_updated_ballet(caplog):
    caplog.set_level(logging.DEBUG, ballet.util.log.logger.name)
    latest = 'x.y.z'
    _warn_of_updated_ballet(latest)
    assert latest in caplog.text


def test_make_template_branch_merge_commit_message():
    result = _make_template_branch_merge_commit_message()
    assert ballet.__version__ in result


def test_safe_delete_remote():
    repo = Mock(spec=git.Repo)
    name = 'name'
    _safe_delete_remote(repo, name)
    repo.delete_remote.assert_called_once_with(name)


def test_log_recommended_reinstall(caplog):
    caplog.set_level(logging.DEBUG, ballet.util.log.logger.name)
    _log_recommended_reinstall()
    assert caplog.text



from typing import Callable, Optional

import funcy as fy
import numpy as np
import pandas as pd
import sklearn.base
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils.validation import check_is_fitted
from sklearn_pandas import DataFrameMapper

import ballet.transformer  # avoid circular import
from ballet.exc import BalletError
from ballet.util import get_arr_desc
from ballet.util.typing import OneOrMore, TransformerLike

__all__ = (
    'BaseTransformer',
    'ConditionalTransformer',
    'GroupedFunctionTransformer',
    'GroupwiseTransformer',
    'NoFitMixin',
    'SimpleFunctionTransformer',
    'SubsetTransformer',
)


class NoFitMixin:
    """Mix-in class for transformations that do not require a fit stage"""

    def fit(self, X, y=None, **fit_kwargs):
        return self


class BaseTransformer(NoFitMixin, TransformerMixin, BaseEstimator):
    """Base transformer class for developing new transformers"""
    pass


class SimpleFunctionTransformer(FunctionTransformer):
    """Transformer that applies a callable to its input

    The callable will be called on the input X in the transform stage,
    optionally with additional arguments and keyword arguments.

    A simple wrapper around :py:class:`FunctionTransformer`.

    Args:
        func: callable to apply
        func_kwargs: keyword arguments to pass
    """

    def __init__(self,
                 func: Callable,
                 func_kwargs: Optional[dict] = None):
        self.func = func
        self.func_kwargs = func_kwargs or {}
        super().__init__(
            func=self.func,
            kw_args=self.func_kwargs)


class GroupedFunctionTransformer(FunctionTransformer):
    """Transformer that applies a callable to each group of a groupby

    Args:
        func: callable to apply
        func_kwargs: keyword arguments to pass
        groupby_kwargs: keyword arguments to ``pd.DataFrame.groupby``. If
            omitted, no grouping is performed and the function is called on
            the entire DataFrame.
    """

    def __init__(self,
                 func: Callable,
                 func_kwargs: Optional[dict] = None,
                 groupby_kwargs: Optional[dict] = None):
        self.func = func
        self.func_kwargs = func_kwargs or {}
        self.groupby_kwargs = groupby_kwargs or {}
        super().__init__(
            func=func,
            kw_args=self.func_kwargs)

    def transform(self, X, **transform_kwargs):
        if self.groupby_kwargs:
            call = X.groupby(**self.groupby_kwargs).apply
        else:
            call = X.pipe
        return call(super().transform)


class GroupwiseTransformer(BaseTransformer):
    """Transformer that does something different for every group

    For each group identified in the training set by the groupby operation,
    a separate transformer is cloned and fit. This is useful to learn
    group-wise transformers that do not leak data between the training and
    test sets. Consider the case of imputing missing values with the mean of
    some group. A normal, pure-pandas implementation, such as
    ``X_te.groupby(by='foo').apply('mean')`` would leak information about
    the test set means, which might differ from the training set means.

    Args:
        transformer: the transformer to apply
            to each group. If transformer is a transformer-like instance (i.e.
            has fit, transform methods etc.), then it is cloned for each group.
            If transformer is a transformer-like class (i.e. instances of
            the class are transformer-like), then it is initialized with no
            arguments for each group. If it is a callable, then it is called
            with no arguments for each group.
        groupby_kwargs: keyword arguments to pd.DataFrame.groupby
        column_selection): column, or list of columns,
            to select after the groupby. Equivalent to
            ``df.groupby(...)[column_selection]``. Defaults to None, i.e. no
            column selection is performed.
        handle_unknown: 'error' or 'ignore', default='error'. Whether to
            raise an error or ignore if an unknown group is encountered during
            transform. When this parameter is set to 'ignore' and an unknown
            group is encountered during transform, the group's values will be
            passed through unchanged.
        handle_error: 'error' or 'ignore', default='error'. Whether to
            raise an error or ignore if an error is raised during transforming
            an individual group. When this parameter is set to 'ignore' and
            an error is raised when calling the transformer's transform
            method on an individual group, the group's values will be passed
            through unchanged.

    Example usage:

        In this example, we create a groupwise transformer that fits a
        separate imputer for each group encountered. For new data points,
        values will be imputed according to the mean of its group on the
        training set, avoiding any data leakage.

        .. code-block:: python

           >>> from sklearn.impute import SimpleImputer
           >>> transformer = GroupwiseTransformer(
           ...     SimpleImputer(strategy='mean'),
           ...     groupby_kwargs = {'level': 'name'}
           ... )

    Raises:
        ballet.exc.BalletError: if handle_unknown=='error' and an unknown group
            is encountered at transform-time.
    """

    def __init__(self,
                 transformer: TransformerLike,
                 groupby_kwargs: dict = None,
                 column_selection: OneOrMore[str] = None,
                 handle_unknown: str = 'error',
                 handle_error: str = 'error'):
        self.satisfy_transformer = transformer
        self.groupby_kwargs = groupby_kwargs
        self.column_selection = column_selection
        self.handle_unknown = handle_unknown
        self.handle_error = handle_error

    def _make_transformer(self):
        if (
            isinstance(self.satisfy_transformer, type)
            or callable(self.satisfy_transformer)
        ):
            return self.satisfy_transformer()
        else:
            return sklearn.base.clone(self.satisfy_transformer)

    def fit(self, X, y=None, **fit_kwargs):
        # validation on inputs
        self.groupby_kwargs_ = self.groupby_kwargs or {}
        if self.handle_unknown not in ['error', 'ignore']:
            raise ValueError(
                f'Invalid value for handle_unknown: {self.handle_unknown}')
        if self.handle_error not in ['error', 'ignore']:
            raise ValueError(
                f'Invalid value for handle_error: {self.handle_error}')

        # Get the groups
        grouper = X.groupby(**self.groupby_kwargs_)
        self.groups_ = set(grouper.groups.keys())

        # Create and fit a transformer for each group
        self.transformers_ = {}
        for group_name, x_group in grouper:
            transformer = self._make_transformer()

            if self.column_selection is not None:
                x_group = x_group[self.column_selection]

            if y is not None:
                # Extract y by integer indexing
                y_group = y[grouper.indices[group_name]]
                transformer.fit(x_group, y_group)
            else:
                transformer.fit(x_group)

            self.transformers_[group_name] = transformer

        return self

    def transform(self, X, **transform_kwargs):
        check_is_fitted(self, ['groups_', 'transformers_'])

        def _transform(x_group, *args, **kwargs):
            # If the group is not a DataFrame, there are two problems
            # 1. We can't rely on group.name to lookup the right transformer
            # 2. We can't "reassemble" the transformed
            # However, the contract of ``pandas.core.groupby.GroupBy.apply`` is
            # that the input is a DataFrame, so this should never occur.
            if not isinstance(x_group, pd.DataFrame):
                raise NotImplementedError

            group_name = x_group.name

            if self.column_selection is not None:
                x_group = x_group[self.column_selection]

            if group_name in self.transformers_:
                transformer = self.transformers_[group_name]
                try:
                    data = transformer.transform(x_group, *args, **kwargs)

                    # This post-processing step is required because sklearn
                    # transform converts a DataFrame to an array. This is my
                    # best attempt so far to approximate the following:
                    # >>> result = x_group.copy()
                    # >>> result.values = data
                    # which is an error as `values` cannot be set.
                    index = x_group.index
                    columns = x_group.columns
                    return pd.DataFrame(
                        data=data, index=index, columns=columns)
                except Exception:
                    if self.handle_error == 'ignore':
                        return x_group
                    else:
                        raise
            else:
                if self.handle_unknown == 'error':
                    raise BalletError(f'Unknown group: {group_name}')
                elif self.handle_unknown == 'ignore':
                    return x_group
                else:
                    # Unreachable code
                    raise RuntimeError

        return (
            X
            .groupby(**self.groupby_kwargs_)
            .apply(_transform, **transform_kwargs)
        )


class ConditionalTransformer(BaseTransformer):
    """Transform columns that satisfy a condition during training

    In the fit stage, determines which variables (columns) satisfy the
    condition. In the transform stage, applies the given transformation to
    the satisfied columns. If a second transformation is given, applies the
    second transformation to the complement of the satisfied columns (i.e.
    the columns that fail to satisfy the condition). Otherwise, these
    unsatisfied columns are passed through unchanged.

    Args:
        condition: condition function
        satisfy_transform: transform function for satisfied columns
        unsatisfy_transform: transform function for unsatisfied columns
            (defaults to identity)
    """

    def __init__(
        self,
        condition: Callable,
        satisfy_transform: Callable,
        unsatisfy_transform: Optional[Callable] = None
    ):
        super().__init__()
        self.condition = condition
        self.satisfy_transform = satisfy_transform
        self.unsatisfy_transform = unsatisfy_transform or fy.identity

    def fit(self, X, y=None, **fit_args):
        # satisfied_columns_ is a bool or array[bool]
        self.satisfied_columns_ = self.condition(X)
        self.unsatisfied_columns_ = np.logical_not(self.satisfied_columns_)
        return self

    def transform(self, X, **transform_args):
        check_is_fitted(self, ['satisfied_columns_', 'unsatisfied_columns_'])

        if isinstance(X, pd.DataFrame):
            X = X.copy()
            X.loc[:, self.satisfied_columns_] = self.satisfy_transform(
                X.loc[:, self.satisfied_columns_])
            X.loc[:, self.unsatisfied_columns_] = self.unsatisfy_transform(
                X.loc[:, self.unsatisfied_columns_]
            )
            return X
        elif np.ndim(X) == 1:
            return (
                self.satisfy_transform(X)
                if self.satisfied_columns_
                else self.unsatisfy_transform(X)
            )
        elif isinstance(X, np.ndarray):
            X = X.copy().astype('float')
            if self.satisfied_columns_.any():
                mask = np.tile(self.satisfied_columns_, (X.shape[0], 1))
                np.putmask(X, mask, self.satisfy_transform(
                    X[:, self.satisfied_columns_]))
            if self.unsatisfied_columns_.any():
                mask = np.tile(self.unsatisfied_columns_, (X.shape[0], 1))
                np.putmask(X, mask, self.unsatisfy_transform(
                    X[:, self.unsatisfied_columns_]))
            return X
        elif not self.satisfied_columns_:
            # if we wouldn't otherwise have known what to do, we can pass
            # through X if transformation was not necessary anyways
            return self.unsatisfy_transform(X)
        else:
            raise TypeError(
                f'Couldn\'t apply transformer on features in '
                f'{get_arr_desc(X)}.')


class SubsetTransformer(DataFrameMapper):
    """Transform a subset of columns with another transformer

    Args:
        input:
        transformer:
        alias:
    """

    def __init__(self,
                 input: OneOrMore[str],
                 transformer: TransformerLike,
                 alias: Optional[str] = None):
        self.input = input
        self.transformer = transformer
        self.alias = alias
        super().__init__(
            [(input,
              ballet.transformer.desugar_transformer(transformer),
              {'alias': alias})],
            default=None,
            input_df=True,
            df_out=True,
        )



import funcy
import pytest

from ballet.eng.misc import IdentityTransformer
from ballet.feature import Feature
from ballet.pipeline import FeatureEngineeringPipeline


@pytest.fixture
def inputs():
    input = 'foo'
    transformer = IdentityTransformer()
    return input, transformer


def test_feature_init_with_one_transformer(inputs):
    input, transformer = inputs
    Feature(input, transformer)


def test_feature_init_with_list_of_transformer(inputs):
    input, transformer = inputs
    Feature(input, [transformer, transformer])


def test_feature_init_with_callable(inputs):
    input, transformer = inputs
    Feature(input, funcy.identity)


def test_feature_init_with_list_of_transformers_and_callables(inputs):
    input, transformer = inputs
    Feature(input, [transformer, funcy.identity])


def test_feature_init_with_none(inputs):
    input, transformer = inputs
    Feature(input, None)


def test_feature_init_with_list_of_none(inputs):
    input, transformer = inputs
    Feature(input, [None, None])


def test_feature_init_with_list_of_none_and_notnone(inputs):
    input, transformer = inputs
    Feature(input, [None, transformer])


def test_feature_init_invalid_transformer_api(inputs):
    input, transformer = inputs
    with pytest.raises(ValueError):
        Feature(input, object())

    with pytest.raises(ValueError):
        Feature(input, IdentityTransformer)


def test_feature_as_input_transformer_tuple(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    tup = feature.as_input_transformer_tuple()
    assert isinstance(tup, tuple)
    assert len(tup) == 3


def test_feature_as_feature_engineering_pipeline(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    mapper = feature.as_feature_engineering_pipeline()
    assert isinstance(mapper, FeatureEngineeringPipeline)


def test_feature_pipeline(inputs):
    input, transformer = inputs
    feature = Feature(input, transformer)
    pipeline = feature.pipeline
    assert isinstance(pipeline, FeatureEngineeringPipeline)
    pipeline2 = feature.pipeline
    assert pipeline is pipeline2



from sklearn.impute import KNNImputer
from sklearn.impute import MissingIndicator
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import Binarizer
from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import KBinsDiscretizer
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import PolynomialFeatures
from sklearn.preprocessing import PowerTransformer
from sklearn.preprocessing import QuantileTransformer
from sklearn.preprocessing import RobustScaler
from sklearn.preprocessing import StandardScaler
from sklearn.random_projection import GaussianRandomProjection
from sklearn.random_projection import SparseRandomProjection


__all__ = [
    'Binarizer',
    'FunctionTransformer',
    'GaussianRandomProjection',
    'KBinsDiscretizer',
    'KNNImputer',
    'MaxAbsScaler',
    'MinMaxScaler',
    'MissingIndicator',
    'Normalizer',
    'OneHotEncoder',
    'OrdinalEncoder',
    'PolynomialFeatures',
    'PowerTransformer',
    'QuantileTransformer',
    'RobustScaler',
    'SimpleImputer',
    'SparseRandomProjection',
    'StandardScaler',
]

try:
    from sklearn.impute import IterativeImputer
    __all__.append('IterativeImputer')
except ImportError:
    pass



import pathlib
from unittest.mock import Mock, patch

import pytest
from funcy import identity

from ballet.util.fs import (
    _synctree, isemptyfile, replaceext, spliceext, splitext2, synctree,)


@pytest.mark.parametrize(
    'convert_path',
    [identity, pathlib.Path],
)
def test_spliceext(convert_path):
    filepath0 = '/foo/bar/baz.txt'

    filepath = convert_path(filepath0)
    s = '_new'
    expected = '/foo/bar/baz_new.txt'

    actual = spliceext(filepath, s)
    assert actual == expected


@pytest.mark.parametrize(
    'convert_path',
    [identity, pathlib.Path],
)
def test_replaceext(convert_path):
    filepath0 = '/foo/bar/baz.txt'
    expected = '/foo/bar/baz.py'

    filepath = convert_path(filepath0)

    new_ext = 'py'
    actual = replaceext(filepath, new_ext)
    assert actual == expected

    new_ext = '.py'
    actual = replaceext(filepath, new_ext)
    assert actual == expected


@pytest.mark.parametrize(
    'convert_path',
    [identity, pathlib.Path],
)
def test_splitext2(convert_path):
    filepath0 = '/foo/bar/baz.txt'
    expected = ('/foo/bar', 'baz', '.txt')

    filepath = convert_path(filepath0)

    actual = splitext2(filepath)
    assert actual == expected

    filepath0 = 'baz.txt'
    expected = ('', 'baz', '.txt')

    filepath = convert_path(filepath0)

    actual = splitext2(filepath)
    assert actual == expected


@patch('os.path.exists')
def test_isemptyfile_does_not_exist(mock_exists):
    mock_exists.return_value = False
    result = isemptyfile(
        '/path/to/file/that/hopefully/does/not/exist')
    assert not result


def test_isemptyfile_is_not_empty(tmp_path):
    # file exists and is not empty - false
    filepath = tmp_path.joinpath('file')
    with filepath.open('w') as f:
        f.write('0')
    result = isemptyfile(filepath)
    assert not result


def test_isemptyfile_is_empty(tmp_path):
    # file exists and is empty - true
    filepath = tmp_path.joinpath('file')
    filepath.touch()
    result = isemptyfile(filepath)
    assert result


def test_synctree(tmp_path):
    src = tmp_path.joinpath('x')
    src.joinpath('a', 'b').mkdir(parents=True)
    src.joinpath('a', 'b', 'only_in_src.txt').touch()
    src.joinpath('a', 'c').mkdir()

    dst = tmp_path.joinpath('y')
    dst.joinpath('a', 'b').mkdir(parents=True)
    dst.joinpath('a', 'only_in_dst.txt').touch()

    # patch here in order to avoid messing up tmp_path stuff
    with patch('pathlib.Path.mkdir') as mock_mkdir, \
            patch('os.unlink') as mock_unlink, \
            patch('os.rmdir') as mock_rmdir, \
            patch('ballet.util.fs.copyfile') as mock_copyfile:
        synctree(src, dst)

    # one call to mkdir, for 'a/c'
    mock_mkdir.assert_called_once_with()

    # one call to copyfile, for 'only_in_src.txt'
    path = ('a', 'b', 'only_in_src.txt')
    mock_copyfile.assert_called_once_with(
        src.joinpath(*path),
        dst.joinpath(*path)
    )

    # no calls to cleanup
    mock_rmdir.assert_not_called()
    mock_unlink.assert_not_called()


@pytest.mark.skip(reason='skipping')
def test__synctree():
    # when src is a directory that exists and dst does not exist,
    # then copytree should be called
    src = Mock(spec=pathlib.Path)
    dst = Mock(spec=pathlib.Path)
    dst.exists.return_value = False
    _synctree(src, dst, lambda x: None)


@pytest.mark.xfail
def test_pwalk():
    raise NotImplementedError



import pathlib
from unittest.mock import Mock

import pytest

from ballet.project import relative_to_contrib
from ballet.validation.project_structure.checks import (
    IsAdditionCheck, IsPythonSourceCheck, ModuleNameCheck,
    RelativeNameDepthCheck, SubpackageNameCheck, WithinContribCheck,)


# @pytest.fixture(scope='module')  # TODO
@pytest.fixture
def project(quickstart):
    yield quickstart.project


def test_relative_to_contrib(project):
    contrib_path = project.config.get('contrib.module_path')
    diff = Mock(b_path=f'{contrib_path}/abc.py')

    expected = pathlib.Path('abc.py')
    actual = relative_to_contrib(diff, project)

    assert actual == expected


def test_is_addition_check(project):
    checker = IsAdditionCheck(project)

    mock_diff = Mock(change_type='A')
    assert checker.do_check(mock_diff)

    mock_diff = Mock(change_type='B')
    assert not checker.do_check(mock_diff)


def test_is_python_source_check(project):
    contrib_path = project.config.get('contrib.module_path')
    checker = IsPythonSourceCheck(project)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.py')
    assert checker.do_check(mock_diff)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.xyz')
    assert not checker.do_check(mock_diff)


def test_within_contrib_check(project):
    contrib_path = project.config.get('contrib.module_path')
    checker = WithinContribCheck(project)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.py')
    assert checker.do_check(mock_diff)

    mock_diff = Mock(b_path='foo/hack.py')
    assert not checker.do_check(mock_diff)


def test_subpackage_name_check(project):
    contrib_path = project.config.get('contrib.module_path')
    checker = SubpackageNameCheck(project)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.py')
    assert checker.do_check(mock_diff)

    mock_diff = Mock(b_path=f'{contrib_path}/bob/feature_1.py')
    assert not checker.do_check(mock_diff)


def test_feature_module_name_check(project):
    contrib_path = project.config.get('contrib.module_path')
    checker = ModuleNameCheck(project)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.py')
    assert checker.do_check(mock_diff)

    bad_paths = [
        f'{contrib_path}/user_bob/foo1.py',
        f'{contrib_path}/user_bob/1.py',
        f'{contrib_path}/user_bob/feature_x-1.py',
    ]
    for path in bad_paths:
        mock_diff = Mock(b_path=path)
        assert not checker.do_check(mock_diff)


def test_relative_name_depth_check(project):
    contrib_path = project.config.get('contrib.module_path')
    checker = RelativeNameDepthCheck(project)

    mock_diff = Mock(b_path=f'{contrib_path}/user_bob/feature_1.py')
    assert checker.do_check(mock_diff)

    mock_diff = Mock(
        b_path=f'{contrib_path}/user_bob/a/b/c/d/feature_1.py')
    assert not checker.do_check(mock_diff)
